#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Methods to support discrete measures
"""
# what about sample_unsafe ?
from mystic.math.stats import *
from mystic.math.samples import *
from mystic.math.integrate import *
from mystic.symbolic import generate_solvers, generate_constraint, solve
from mystic.symbolic import generate_conditions, generate_penalty
from mystic.math import almostEqual
from mystic.math.distance import Lnorm

__all__ = ['weighted_select','spread','norm','maximum','ess_maximum',\
          'minimum','ess_minimum','ptp','ess_ptp','expectation',\
          'expected_variance','expected_std','_expected_moment','mean',\
          'support_index','support','moment','standard_moment','variance',\
          'std','skewness','kurtosis','impose_mean','impose_variance',\
          'impose_std','impose_moment','impose_spread','impose_expectation',\
          '_impose_expected_moment','impose_expected_variance',\
          'impose_expected_std','impose_expected_mean_and_variance',\
          'impose_weight_norm','normalize','impose_reweighted_mean',\
          'impose_reweighted_variance','impose_reweighted_std','_sort',\
          'median','mad','impose_median','impose_mad','_k','tmean',\
          'tvariance','tstd','impose_tmean','impose_tvariance','impose_tstd',\
          'impose_support','impose_unweighted','impose_collapse',\
          'impose_sum','impose_product','_pack','_unpack','_flat',\
          '_nested','_nested_split','split_param']

def weighted_select(samples, weights, mass=1.0):
  """randomly select a sample from weighted set of samples

Args:
    samples (list): a list of sample points
    weights (list): a list of sample weights
    mass (float, default=1.0): sum of normalized weights

Returns:
    a randomly selected sample point
"""
  from numpy import sum, array
  from mystic.tools import random_state
  rand = random_state().random
  # generate a list representing the weighted distribution
  wts = normalize(weights, mass)
  wts = array([sum(wts[:i+1]) for i in range(len(wts))])
  # correct for any rounding error
  wts[-1] = mass
  # generate a random weight
  w = mass * rand()
  # select samples that corresponds to randomly selected weight
  selected = len(wts[ wts <= w ])
  return samples[selected]

##### calculate methods #####
def spread(samples):
  """calculate the range for a list of points 

``spread(x) = max(x) - min(x)``

Args:
    samples (list): a list of sample points

Returns:
    the range of the samples
"""
  return max(samples) - min(samples)

def norm(weights):
  """calculate the norm of a list of points

``norm(x) = mean(x)``

Args:
    weights (list): a list of sample weights

Returns:
    the mean of the weights
"""
  return mean(weights)

def maximum(f, samples):
  """calculate the max of function for the given list of points

``maximum(f,x) = max(f(x))``

Args:
    f (func): a function that takes a list and returns a number
    samples (list): a list of sample points

Returns:
    the maximum output value for a function at the given inputs
"""
  y = [f(x) for x in samples] #XXX: parallel map?
  return max(y)

def ess_maximum(f, samples, weights=None, tol=0.):
  """calculate the max of function for support on the given list of points

``ess_maximum(f,x,w) = max(f(support(x,w)))``

Args:
    f (func): a function that takes a list and returns a number
    samples (list): a list of sample points
    weights (list, default=None): a list of sample weights
    tol (float, default=0.0): a tolerance, where any ``weight <= tol`` is zero

Returns:
    the maximum output value for a function at the given support points
"""
  if weights is None:
    return maximum(f, samples)
  return maximum(f, support(samples, weights, tol))

def minimum(f, samples):
  """calculate the min of function for the given list of points

``minimum(f,x) = min(f(x))``

Args:
    f (func): a function that takes a list and returns a number
    samples (list): a list of sample points

Returns:
    the minimum output value for a function at the given inputs
"""
  y = [f(x) for x in samples] #XXX: parallel map?
  return min(y)

def ess_minimum(f, samples, weights=None, tol=0.):
  """calculate the min of function for support on the given list of points

``ess_minimum(f,x,w) = min(f(support(x,w)))``

Args:
    f (func): a function that takes a list and returns a number
    samples (list): a list of sample points
    weights (list, default=None): a list of sample weights
    tol (float, default=0.0): a tolerance, where any ``weight <= tol`` is zero

Returns:
    the minimum output value for a function at the given support points
"""
  if weights is None:
    return minimum(f, samples)
  return minimum(f, support(samples, weights, tol))

def ptp(f, samples):
  """calculate the spread of function for the given list of points

``minimum(f,x) = max(f(x)) - min(f(x))``

Args:
    f (func): a function that takes a list and returns a number
    samples (list): a list of sample points

Returns:
    the spread in output value for a function at the given inputs
"""
  y = [f(x) for x in samples] #XXX: parallel map?
  return max(y) - min(y)

def ess_ptp(f, samples, weights=None, tol=0.):
  """calculate the spread of function for support on the given list of points

``ess_minimum(f,x,w) = max(f(support(x,w))) - min(f(support(x,w)))``

Args:
    f (func): a function that takes a list and returns a number
    samples (list): a list of sample points
    weights (list, default=None): a list of sample weights
    tol (float, default=0.0): a tolerance, where any ``weight <= tol`` is zero

Returns:
    the spread in output value for a function at the given support points
"""
  if weights is None:
    return ptp(f, samples)
  return ptp(f, support(samples, weights, tol))

def expectation(f, samples, weights=None, tol=0.0):
  """calculate the (weighted) expectation of a function for a list of points

Args:
    f (func): a function that takes a list and returns a number
    samples (list): a list of sample points
    weights (list, default=None): a list of sample weights
    tol (float, default=0.0): a tolerance, where any ``weight <= tol`` is zero

Returns:
    the weighted expectation for a list of sample points
"""
  if weights is None:
    y = [f(x) for x in samples] #XXX: parallel map?
    return mean(y, weights)
  # contributed by TJS #
  # to prevent function evaluation if weight is "too small":
  # skip evaluation of f(x) if the corresponding weight <= tol
  # import itertools as it   #XXX: assumes is faster than zip
  #weights = normalize(weights, mass=1.0) #FIXME: below is atol, should be rtol?
  if not sum(abs(w) > tol for w in weights):
      yw = ((0.0,0.0),)
  else: #XXX: parallel map?
      yw = [(f(x),w) for (x,w) in zip(samples, weights) if abs(w) > tol]
  return mean(*zip(*yw))
  ##XXX: at around len(samples) == 150, the following is faster
  #aw = asarray(weights)
  #ax = asarray(samples)
  #w = aw > tol
  #return mean(ax[w], aw[w])

def _expected_moment(f, samples, weights=None, order=1, tol=0.0):
  """calculate the (weighted) nth-order expected moment of a function

Args:
    f (func): a function that takes a list and returns a number
    samples (list): a list of sample points
    weights (list, default=None): a list of sample weights
    order (int, default=1): the degree, a positive integer
    tol (float, default=0.0): a tolerance, where any ``weight <= tol`` is zero

Returns:
    the weighted nth-order expected moment of f on a list of sample points
"""
  if order < 0:
    raise NotImplementedError
  if weights is None:
    y = [f(x) for x in samples] #XXX: parallel map?
    return moment(y, weights, order) #XXX: tol?
  # skip evaluation of f(x) if the corresponding weight <= tol
  if not sum(abs(w) > tol for w in weights):
      yw = ((0.0,0.0),)
  else: #XXX: parallel map?
      yw = [(f(x),w) for (x,w) in zip(samples, weights) if abs(w) > tol]
  return moment(*zip(*yw), order=order)


def expected_variance(f, samples, weights=None, tol=0.0):
  """calculate the (weighted) expected variance of a function

Args:
    f (func): a function that takes a list and returns a number
    samples (list): a list of sample points
    weights (list, default=None): a list of sample weights
    tol (float, default=0.0): a tolerance, where any ``weight <= tol`` is zero

Returns:
    the weighted expected variance of f on a list of sample points
"""
  return _expected_moment(f, samples, weights, order=2, tol=tol)


def expected_std(f, samples, weights=None, tol=0.0):
  """calculate the (weighted) expected standard deviation of a function

Args:
    f (func): a function that takes a list and returns a number
    samples (list): a list of sample points
    weights (list, default=None): a list of sample weights
    tol (float, default=0.0): a tolerance, where any ``weight <= tol`` is zero

Returns:
    the weighted expected standard deviation of f on a list of sample points
"""
  from numpy import sqrt
  return sqrt(expected_variance(f, samples, weights, tol=tol))


def mean(samples, weights=None, tol=0):
  """calculate the (weighted) mean for a list of points

Args:
    samples (list): a list of sample points
    weights (list, default=None): a list of sample weights
    tol (float, default=0.0): a tolerance, where any ``mean <= tol`` is zero

Returns:
    the weighted mean for a list of sample points
"""
  if weights is None:
    weights = [1.0/float(len(samples))] * len(samples)
  # get weighted sum
  ssum = sum(i*j for i,j in zip(samples, weights))
  # normalize by sum of the weights
  wts = float(sum(weights))
  if wts:
    ssum = ssum / wts
    return 0.0 if abs(ssum) <= tol else ssum
  from numpy import inf
  return ssum * inf  # protect against ZeroDivision

def support_index(weights, tol=0): #XXX: no relative tolerance near zero
  """get the indices of the positions which have non-zero weight

Args:
    weights (list): a list of sample weights
    tol (float, default=0.0): a tolerance, where any ``weight <= tol`` is zero

Returns:
    a list of indices of positions with non-zero weight
"""
  return [i for (i,w) in enumerate(weights) if w > tol]

def support(samples, weights, tol=0): #XXX: no relative tolerance near zero
  """get the positions which have non-zero weight

Args:
    samples (list): a list of sample points
    weights (list): a list of sample weights
    tol (float, default=0.0): a tolerance, where any ``weight <= tol`` is zero

Returns:
    a list of positions with non-zero weight
"""
  return [samples[i] for (i,w) in enumerate(weights) if w > tol]

def moment(samples, weights=None, order=1, tol=0): #, _mean=None):
  """calculate the (weighted) nth-order moment for a list of points

Args:
    samples (list): a list of sample points
    weights (list, default=None): a list of sample weights
    order (int, default=1): the degree, a positive integer
    tol (float, default=0.0): a tolerance, where any ``mean <= tol`` is zero

Returns:
    the weighted nth-order moment for a list of sample points
"""
  if order == 0: return 1.0 #XXX: error if order < 0
  if order == 1: return 0.0
  if weights is None:
    weights = [1.0/float(len(samples))] * len(samples)
 #if _mean is None:
  _mean = mean(samples, weights)
  mom = [(s - _mean)**order for s in samples] #XXX: abs(s - _mean) ???
  return mean(mom,weights,tol) #XXX: sample_moment (Bessel correction) *N/(N-1)?

def standard_moment(samples, weights=None, order=1, tol=0):
  """calculate the (weighted) nth-order standard moment for a list of points

``standard_moment(x,w,order) = moment(x,w,order)/std(x,w)^order``

Args:
    samples (list): a list of sample points
    weights (list, default=None): a list of sample weights
    order (int, default=1): the degree, a positive integer
    tol (float, default=0.0): a tolerance, where any ``mean <= tol`` is zero

Returns:
    the weighted nth-order standard moment for a list of sample points
"""
  if order == 2: return 1.0 #XXX: error if order < 0
  return moment(samples, weights, order, tol)/std(samples, weights)**order

def variance(samples, weights=None): #,tol=0, _mean=None):
  """calculate the (weighted) variance for a list of points

Args:
    samples (list): a list of sample points
    weights (list, default=None): a list of sample weights

Returns:
    the weighted variance for a list of sample points
"""
  return moment(samples, weights, order=2)

def std(samples, weights=None): #,tol=0, _mean=None):
  """calculate the (weighted) standard deviation for a list of points

Args:
    samples (list): a list of sample points
    weights (list, default=None): a list of sample weights

Returns:
    the weighted standard deviation for a list of sample points
"""
  from numpy import sqrt
  return sqrt(variance(samples, weights)) # _mean)

def skewness(samples, weights=None): #,tol=0, _mean=None):
  """calculate the (weighted) skewness for a list of points

Args:
    samples (list): a list of sample points
    weights (list, default=None): a list of sample weights

Returns:
    the weighted skewness for a list of sample points
"""
  return standard_moment(samples, weights, order=3) # _mean)

def kurtosis(samples, weights=None): #,tol=0, _mean=None):
  """calculate the (weighted) kurtosis for a list of points

Args:
    samples (list): a list of sample points
    weights (list, default=None): a list of sample weights

Returns:
    the weighted kurtosis for a list of sample points
"""
  return standard_moment(samples, weights, order=4) # _mean)


##### coordinate shift methods #####
from numpy import asarray
def impose_mean(m, samples, weights=None): #,tol=0):
  """impose a mean on a list of (weighted) points

Args:
    m (float): the target mean
    samples (list): a list of sample points
    weights (list, default=None): a list of sample weights

Returns:
    a list of sample points with the desired weighted mean

Notes:
    this function does not alter the weighted range or the weighted variance
"""
 #XXX: this is as expected... mean(impose_mean(2.0, samples, weights), weights)
 #XXX: this is unexpected?... mean(impose_mean(2.0, samples, weights))
  samples = asarray(list(samples)) #XXX: faster to use x = array(x, copy=True) ?
  shift = m - mean(samples, weights)
  samples = samples + shift  #NOTE: is "range-preserving"
  return list(samples)


def impose_variance(v, samples, weights=None): #,tol=0):
  """impose a variance on a list of (weighted) points

Args:
    v (float): the target variance
    samples (list): a list of sample points
    weights (list, default=None): a list of sample weights

Returns:
    a list of sample points with the desired weighted variance

Notes:
    this function does not alter the weighted mean
"""
  m = mean(samples, weights)
  samples = asarray(list(samples)) #XXX: faster to use x = array(x, copy=True) ?
  sv = variance(samples,weights) #,m)
  if not sv:  # protect against ZeroDivision when variance = 0
    if not v: # variance is to be 0
      return [float(i) for i in samples] #samples.tolist()
    from numpy import nan
    return [nan]*len(samples) #XXX: better to space pts evenly across range?
  from numpy import sqrt, seterr
  err = seterr(invalid='ignore')
  scale = sqrt(float(v) / sv)
  seterr(**err)
  samples = samples * scale  #NOTE: not "mean-preserving", until the next line
  return impose_mean(m, samples, weights) #NOTE: not range preserving

#FIXME: for range and variance to be 'mutually preserving'...
#       must reconcile scaling by sqrt(v2/v1) & (r2/r1)
#       ...so likely, must scale the weights... or scale each point differently

def impose_std(s, samples, weights=None):
  """impose a standard deviation on a list of (weighted) points

Args:
    s (float): the target standard deviation
    samples (list): a list of sample points
    weights (list, default=None): a list of sample weights

Returns:
    a list of sample points with the desired weighted standard deviation

Notes:
    this function does not alter the weighted mean
"""
  return impose_variance(s**2, samples, weights)


def impose_moment(m, samples, weights=None, order=1, tol=0, skew=None):
  """impose the selected moment on a list of (weighted) points

Args:
    m (float): the target moment
    samples (list): a list of sample points
    weights (list, default=None): a list of sample weights
    order (int, default=1): the degree, a positive integer
    tol (float, default=0.0): a tolerance, where any ``mean <= tol`` is zero
    skew (bool, default=None): if True, allow skew in the samples

Returns:
    a list of sample points with the desired weighted moment

Notes:
    this function does not alter the weighted mean

    if *skew* is None, then allow *skew* when *order* is odd
"""
  v = m #NOTE: change of variables, so code is consistent with impose_variance
  if order == 0:
    if v == 1: return [float(i) for i in samples]
    else:
      from numpy import nan
      return [nan]*len(samples)
  if order == 1:
    if not v: return [float(i) for i in samples]
    else:
      from numpy import nan
      return [nan]*len(samples)
  if not order%2 and v < 0.0:
    from numpy import nan
    return [nan]*len(samples)
  m = mean(samples, weights)
  if skew is None: skew = order%2 # if odd
  if skew: samples = [i**2 for i in samples]
  sv = moment(samples,weights,order,tol)
  if not sv: #moment == 0     #NOTE: caution if ~0.0 or -v
    if not order%2: # not odd
      return [m]*len(samples)
    if not v: # moment is to be 0
      return [float(i) for i in samples] #samples.tolist()
   #if skew: then unskewed should have worked. XXX: handle this?
    from numpy import nan
    return [nan]*len(samples)
  samples = asarray(list(samples))
  from numpy import power, flipud
  fact = float(v)/sv
  # temporarily account for negative fact
  if order%2 and fact < 0: flip = True
  else: flip = False
 #try:
  scale = power(abs(fact), 1./order)
 #except ZeroDivisionError:
 #  from numpy import nan
 #  return [nan]*len(samples)
  if flip: samples = max(samples) + min(samples) - samples
 #if flip: samples = flipud(max(samples) + min(samples) - samples)
  samples = samples * scale
  return impose_mean(m, samples, weights)


def impose_spread(r, samples, weights=None): #FIXME: fails if len(samples) = 1
  """impose a range on a list of (weighted) points

Args:
    r (float): the target range
    samples (list): a list of sample points
    weights (list, default=None): a list of sample weights

Returns:
    a list of sample points with the desired weighted range

Notes:
    this function does not alter the weighted mean
"""
  m = mean(samples, weights)
  samples = asarray(list(samples)) #XXX: faster to use x = array(x, copy=True) ?
  sr = spread(samples)
  if not sr:  # protect against ZeroDivision when range = 0
    from numpy import nan
    return [nan]*len(samples) #XXX: better to space pts evenly across range?
  scale = float(r) / sr
  samples = samples * scale  #NOTE: not "mean-preserving", until the next line
  return impose_mean(m, samples, weights) #NOTE: not variance preserving


#NOTE: backward incompatible 08/26/18: (param=(m,D),...) --> (m,...,tol=D) 
def impose_expectation(m, f, npts, bounds=None, weights=None, **kwds):
  """impose a given expectation value ``E`` on a given function *f*,
where ``E = m +/- tol`` and ``E = mean(f(x))`` for ``x`` in *bounds*

Args:
    m (float): target expected mean
    f (func): a function that takes a list and returns a number
    npts (tuple(int)): a tuple of dimensions of the target product measure
    bounds (tuple, default=None): tuple is ``(lower_bounds, upper_bounds)``
    weights (list, default=None): a list of sample weights
    tol (float, default=None): maximum allowable deviation from ``m``
    constraints (func, default=None): a function that takes a nested list of
        ``N x 1D`` discrete measure positions and weights, with the intended
        purpose of kernel-transforming ``x,w`` as ``x' = constraints(x, w)``
    npop (int, default=200): size of the trial solution population
    maxiter (int, default=1000): the maximum number of iterations to perform
    maxfun (int, default=1e+6): the maximum number of function evaluations

Returns:
    a list of sample positions, with expectation ``E``

Notes:
    Expectation value ``E`` is calculated by minimizing ``mean(f(x)) - m``,
    over the given *bounds*, and will terminate when ``E`` is found within
    deviation ``tol`` of the target mean ``m``.  If ``tol`` is not provided,
    then a relative deviation of 1% of ``m`` will be used.

    This function does not preserve the mean, variance, or range, as there
    is no initial list of samples to draw the mean, variance, and etc from

    *bounds* is tuple with ``length(bounds) == 2``, composed of all the lower
    bounds, then all the upper bounds, for each parameter

Examples:
    >>> # provide the dimensions and bounds
    >>> nx = 3;  ny = 2;  nz = 1
    >>> x_lb = [10.0];  y_lb = [0.0];  z_lb = [10.0]
    >>> x_ub = [50.0];  y_ub = [9.0];  z_ub = [90.0]
    >>> 
    >>> # prepare the bounds
    >>> lb = (nx * x_lb) + (ny * y_lb) + (nz * z_lb)
    >>> ub = (nx * x_ub) + (ny * y_ub) + (nz * z_ub)
    >>>
    >>> # generate a list of samples with mean +/- dev imposed
    >>> mean = 2.0;  dev = 0.01
    >>> samples = impose_expectation(mean, f, (nx,ny,nz), (lb,ub), tol=dev)
    >>>
    >>> # test the results by calculating the expectation value for the samples
    >>> expectation(f, samples)
    >>> 2.000010010122465
"""
  # param[0] is the target mean
  # param[1] is the acceptable deviation from the target mean
  tol = kwds['tol'] if 'tol' in kwds else None
  param = (m,tol or 0.01*m)

  # FIXME: the following is a HACK to recover from lost 'weights' information
  #        we 'mimic' discrete measures using the product measure weights
  # plug in the 'constraints' function:  samples' = constrain(samples, weights)
  constrain = None   # default is no constraints
  if 'constraints' in kwds: constrain = kwds['constraints']
  if not constrain:  # if None (default), there are no constraints
    constraints = lambda x: x
  else: #XXX: better to use a standard "xk' = constrain(xk)" interface ?
    def constraints(rv):
      coords = _pack( _nested(rv,npts) )
      coords = list(zip(*coords))              # 'mimic' a nested list
      coords = constrain(coords, [weights for i in range(len(coords))])
      coords = list(zip(*coords))              # revert back to a packed list
      return _flat( _unpack(coords,npts) )

  # construct cost function to reduce deviation from expectation value
  def cost(rv):
    """compute cost from a 1-d array of model parameters,
    where:  cost = | E[model] - m |**2 """
    # from mystic.math.measures import _pack, _nested, expectation
    samples = _pack( _nested(rv,npts) )
    Ex = expectation(f, samples, weights)
    return (Ex - param[0])**2

  # if bounds are not set, use the default optimizer bounds
  if not bounds:
    lower_bounds = []; upper_bounds = []
    for n in npts:
      lower_bounds += [None]*n
      upper_bounds += [None]*n
  else: 
    lower_bounds, upper_bounds = bounds

  # construct and configure optimizer
  debug = kwds['debug'] if 'debug' in kwds else False
  npop = kwds.pop('npop', 200)
  maxiter = kwds.pop('maxiter', 1000)
  maxfun = kwds.pop('maxfun', 1e+6)
  crossover = 0.9; percent_change = 0.9

  def optimize(cost, bounds, tolerance, _constraints):
    (lb,ub) = bounds
    from mystic.solvers import DifferentialEvolutionSolver2
    from mystic.termination import VTR
    from mystic.strategy import Best1Exp
    from mystic.monitors import VerboseMonitor, Monitor
    from mystic.tools import random_seed
    if debug: random_seed(123)
    evalmon = Monitor();  stepmon = Monitor()
    if debug: stepmon = VerboseMonitor(10)

    ndim = len(lb)
    solver = DifferentialEvolutionSolver2(ndim,npop)
    solver.SetRandomInitialPoints(min=lb,max=ub)
    solver.SetStrictRanges(min=lb,max=ub)
    solver.SetEvaluationLimits(maxiter,maxfun)
    solver.SetEvaluationMonitor(evalmon)
    solver.SetGenerationMonitor(stepmon)
    solver.Solve(cost,termination=VTR(tolerance),strategy=Best1Exp, \
                 CrossProbability=crossover,ScalingFactor=percent_change, \
                 constraints = _constraints) #XXX: parallel map?

    solved = solver.Solution()
    diameter_squared = solver.bestEnergy
    func_evals = len(evalmon)
    return solved, diameter_squared, func_evals

  # use optimization to get expectation value
  tolerance = (param[1])**2
  results = optimize(cost, (lower_bounds, upper_bounds), tolerance, constraints)

  # repack the results
  samples = _pack( _nested(results[0],npts) )
  return samples
 

def _impose_expected_moment(m, f, npts, bounds=None, weights=None, **kwds):
  """impose a given expected moment ``E`` on a given function *f*,
where ``E = m +/- tol`` and ``E = moment(f(x))`` for ``x`` in *bounds*

Args:
    m (float): target expected moment
    f (func): a function that takes a list and returns a number
    npts (tuple(int)): a tuple of dimensions of the target product measure
    bounds (tuple, default=None): tuple is ``(lower_bounds, upper_bounds)``
    weights (list, default=None):  a list of sample weights
    order (int, default=1): the degree, a positive integer
    tol (float, default=None): maximum allowable deviation from ``m``
    constraints (func, default=None): a function that takes a nested list of
        ``N x 1D`` discrete measure positions and weights, with the intended
        purpose of kernel-transforming ``x,w`` as ``x' = constraints(x, w)``
    npop (int, default=200): size of the trial solution population
    maxiter (int, default=1000): the maximum number of iterations to perform
    maxfun (int, default=1e+6): the maximum number of function evaluations

Returns:
    a list of sample positions, with expected moment ``E``

Notes:
    Expected moment ``E`` is calculated by minimizing ``moment(f(x)) - m``,
    over the given *bounds*, and will terminate when ``E`` is found within
    deviation ``tol`` of the target moment ``m``. If ``tol`` is not provided,
    then a relative deviation of 1% of ``m`` will be used.

    This function does not preserve the mean, variance, or range, as there
    is no initial list of samples to draw the mean, variance, and etc from

    *bounds* is tuple with ``length(bounds) == 2``, composed of all the lower
    bounds, then all the upper bounds, for each parameter

Examples:
    >>> # provide the dimensions and bounds
    >>> nx = 3;  ny = 2;  nz = 1
    >>> x_lb = [10.0];  y_lb = [0.0];  z_lb = [10.0]
    >>> x_ub = [50.0];  y_ub = [9.0];  z_ub = [90.0]
    >>> 
    >>> # prepare the bounds
    >>> lb = (nx * x_lb) + (ny * y_lb) + (nz * z_lb)
    >>> ub = (nx * x_ub) + (ny * y_ub) + (nz * z_ub)
    >>>
    >>> # generate a list of samples with moment +/- dev imposed
    >>> mom = 2.0;  dev = 0.01
    >>> samples = _impose_expected_moment(mom, f, (nx,ny,nz), (lb,ub), \
    ...                                            order=2, tol=dev)
    >>>
    >>> # test the results by calculating the expected moment for the samples
    >>> _expected_moment(f, samples, order=2)
    >>> 2.000010010122465
"""
  # param[0] is the target moment (i.e. mean, variance, ...)
  # param[1] is the acceptable deviation from the target moment
  # param[2] is the order of the moment (i.e. 2=variance, ...)
  tol = kwds['tol'] if 'tol' in kwds else None
  order = kwds['order'] if 'order' in kwds else 1
  param = (m, tol or 0.01*m, order)

  if param[-1] < 0:
     msg = 'order must be greater than zero (order = %s)' % param[-1]
     raise ValueError(msg)
  if param[-1] == 1 and not (param[1] >= -param[0] >= -param[1]):
     msg = 'if order == 1, then moment == 0 (moment = %s +/- %s)' % param[:-1]
     raise ValueError(msg)
  if param[-1] == 0 and not (param[1] >= 1-param[0] >= -param[1]):
     msg = 'if order == 0, then moment == 1 (moment %s +/- %s)' % param[:-1]
     raise ValueError(msg)

  # FIXME: the following is a HACK to recover from lost 'weights' information
  #        we 'mimic' discrete measures using the product measure weights
  # plug in the 'constraints' function:  samples' = constrain(samples, weights)
  constrain = None   # default is no constraints
  if 'constraints' in kwds: constrain = kwds['constraints']
  if not constrain:  # if None (default), there are no constraints
    constraints = lambda x: x
  else: #XXX: better to use a standard "xk' = constrain(xk)" interface ?
    def constraints(rv):
      coords = _pack( _nested(rv,npts) )
      coords = list(zip(*coords))              # 'mimic' a nested list
      coords = constrain(coords, [weights for i in range(len(coords))])
      coords = list(zip(*coords))              # revert back to a packed list
      return _flat( _unpack(coords,npts) )

  # construct cost function to reduce deviation from expected moment
  def cost(rv):
    """compute cost from a 1-d array of model parameters,
    where:  cost = | E_moment[model] - m |**2 """
    # from mystic.math.measures import _pack, _nested, _expected_moment
    samples = _pack( _nested(rv,npts) )
    Ex = _expected_moment(f, samples, weights, param[-1])
    return (Ex - param[0])**2

  # if bounds are not set, use the default optimizer bounds
  if not bounds:
    lower_bounds = []; upper_bounds = []
    for n in npts:
      lower_bounds += [None]*n
      upper_bounds += [None]*n
  else:
    lower_bounds, upper_bounds = bounds

  if len(lower_bounds) == 1 and param[-1] > 1 \
     and not (param[1] >= -param[0] >= -param[1]):
     msg = 'if size == 1, then moment == 0 (moment %s +/- %s)' % param[:-1]
     raise ValueError(msg)

  # construct and configure optimizer
  debug = kwds['debug'] if 'debug' in kwds else False
  npop = kwds.pop('npop', 200)
  maxiter = 1 if param[-1] <= 1 else 1000
  maxiter = kwds.pop('maxiter', maxiter)
  maxfun = kwds.pop('maxfun', 1e+6)
  crossover = 0.9; percent_change = 0.9

  def optimize(cost, bounds, tolerance, _constraints):
    (lb,ub) = bounds
    from mystic.solvers import DifferentialEvolutionSolver2
    from mystic.termination import VTR
    from mystic.strategy import Best1Exp
    from mystic.monitors import VerboseMonitor, Monitor
    from mystic.tools import random_seed
    if debug: random_seed(123)
    evalmon = Monitor();  stepmon = Monitor()
    if debug: stepmon = VerboseMonitor(10)

    ndim = len(lb)
    solver = DifferentialEvolutionSolver2(ndim,npop)
    solver.SetRandomInitialPoints(min=lb,max=ub)
    solver.SetStrictRanges(min=lb,max=ub)
    solver.SetEvaluationLimits(maxiter,maxfun)
    solver.SetEvaluationMonitor(evalmon)
    solver.SetGenerationMonitor(stepmon)
    solver.Solve(cost,termination=VTR(tolerance),strategy=Best1Exp, \
                 CrossProbability=crossover,ScalingFactor=percent_change, \
                 constraints = _constraints) #XXX: parallel map?

    solved = solver.Solution()
    diameter_squared = solver.bestEnergy
    func_evals = len(evalmon)
    return solved, diameter_squared, func_evals

  # use optimization to get expected moment
  tolerance = (param[1])**2
  results = optimize(cost, (lower_bounds, upper_bounds), tolerance, constraints)

  # repack the results
  samples = _pack( _nested(results[0],npts) )
  return samples


def impose_expected_variance(v, f, npts, bounds=None, weights=None, **kwds):
  """impose a given expected variance ``E`` on a given function *f*,
where ``E = v +/- tol`` and ``E = variance(f(x))`` for ``x`` in *bounds*

Args:
    v (float): target expected variance
    f (func): a function that takes a list and returns a number
    npts (tuple(int)): a tuple of dimensions of the target product measure
    bounds (tuple, default=None): tuple is ``(lower_bounds, upper_bounds)``
    weights (list, default=None):  a list of sample weights
    tol (float, default=None): maximum allowable deviation from ``v``
    constraints (func, default=None): a function that takes a nested list of
        ``N x 1D`` discrete measure positions and weights, with the intended
        purpose of kernel-transforming ``x,w`` as ``x' = constraints(x, w)``
    npop (int, default=200): size of the trial solution population
    maxiter (int, default=1000): the maximum number of iterations to perform
    maxfun (int, default=1e+6): the maximum number of function evaluations

Returns:
    a list of sample positions, with expected variance ``E``

Notes:
    Expected variance ``E`` is calculated by minimizing ``variance(f(x)) - v``,
    over the given *bounds*, and will terminate when ``E`` is found within
    deviation ``tol`` of the target variance ``v``. If ``tol`` is not provided,
    then a relative deviation of 1% of ``v`` will be used.

    This function does not preserve the mean, variance, or range, as there
    is no initial list of samples to draw the mean, variance, and etc from

    *bounds* is tuple with ``length(bounds) == 2``, composed of all the lower
    bounds, then all the upper bounds, for each parameter

Examples:
    >>> # provide the dimensions and bounds
    >>> nx = 3;  ny = 2;  nz = 1
    >>> x_lb = [10.0];  y_lb = [0.0];  z_lb = [10.0]
    >>> x_ub = [50.0];  y_ub = [9.0];  z_ub = [90.0]
    >>> 
    >>> # prepare the bounds
    >>> lb = (nx * x_lb) + (ny * y_lb) + (nz * z_lb)
    >>> ub = (nx * x_ub) + (ny * y_ub) + (nz * z_ub)
    >>>
    >>> # generate a list of samples with variance +/- dev imposed
    >>> var = 2.0;  dev = 0.01
    >>> samples = impose_expected_variance(var, f, (nx,ny,nz), (lb,ub), tol=dev)
    >>>
    >>> # test the results by calculating the expected variance for the samples
    >>> expected_variance(f, samples)
    >>> 2.000010010122465
"""
  # param[0] is the target variance
  # param[1] is the acceptable deviation from the target variance
  tol = kwds['tol'] if 'tol' in kwds else None
  param = (v, tol or 0.01*v)

  # FIXME: the following is a HACK to recover from lost 'weights' information
  #        we 'mimic' discrete measures using the product measure weights
  # plug in the 'constraints' function:  samples' = constrain(samples, weights)
  constrain = None   # default is no constraints
  if 'constraints' in kwds: constrain = kwds['constraints']
  if not constrain:  # if None (default), there are no constraints
    constraints = lambda x: x
  else: #XXX: better to use a standard "xk' = constrain(xk)" interface ?
    def constraints(rv):
      coords = _pack( _nested(rv,npts) )
      coords = list(zip(*coords))              # 'mimic' a nested list
      coords = constrain(coords, [weights for i in range(len(coords))])
      coords = list(zip(*coords))              # revert back to a packed list
      return _flat( _unpack(coords,npts) )

  # construct cost function to reduce deviation from expected variance
  def cost(rv):
    """compute cost from a 1-d array of model parameters,
    where:  cost = | E_var[model] - m |**2 """
    # from mystic.math.measures import _pack, _nested, expected_variance
    samples = _pack( _nested(rv,npts) )
    Ex = expected_variance(f, samples, weights)
    return (Ex - param[0])**2

  # if bounds are not set, use the default optimizer bounds
  if not bounds:
    lower_bounds = []; upper_bounds = []
    for n in npts:
      lower_bounds += [None]*n
      upper_bounds += [None]*n
  else:
    lower_bounds, upper_bounds = bounds

  if len(lower_bounds) == 1 and not (param[1] >= -param[0] >= -param[1]):
     msg = 'if size == 1, then variance == 0 (variance %s +/- %s)' % param
     raise ValueError(msg)

  # construct and configure optimizer
  debug = kwds['debug'] if 'debug' in kwds else False
  npop = kwds.pop('npop', 200)
  maxiter = kwds.pop('maxiter', 1000)
  maxfun = kwds.pop('maxfun', 1e+6)
  crossover = 0.9; percent_change = 0.9

  def optimize(cost, bounds, tolerance, _constraints):
    (lb,ub) = bounds
    from mystic.solvers import DifferentialEvolutionSolver2
    from mystic.termination import VTR
    from mystic.strategy import Best1Exp
    from mystic.monitors import VerboseMonitor, Monitor
    from mystic.tools import random_seed
    if debug: random_seed(123)
    evalmon = Monitor();  stepmon = Monitor()
    if debug: stepmon = VerboseMonitor(10)

    ndim = len(lb)
    solver = DifferentialEvolutionSolver2(ndim,npop)
    solver.SetRandomInitialPoints(min=lb,max=ub)
    solver.SetStrictRanges(min=lb,max=ub)
    solver.SetEvaluationLimits(maxiter,maxfun)
    solver.SetEvaluationMonitor(evalmon)
    solver.SetGenerationMonitor(stepmon)
    solver.Solve(cost,termination=VTR(tolerance),strategy=Best1Exp, \
                 CrossProbability=crossover,ScalingFactor=percent_change, \
                 constraints = _constraints) #XXX: parallel map?

    solved = solver.Solution()
    diameter_squared = solver.bestEnergy
    func_evals = len(evalmon)
    return solved, diameter_squared, func_evals

  # use optimization to get expected_variance
  tolerance = (param[1])**2
  results = optimize(cost, (lower_bounds, upper_bounds), tolerance, constraints)

  # repack the results
  samples = _pack( _nested(results[0],npts) )
  return samples


def impose_expected_std(s, f, npts, bounds=None, weights=None, **kwds):
  """impose a given expected std ``E`` on a given function *f*,
where ``E = s +/- tol`` and ``E = std(f(x))`` for ``x`` in *bounds*

Args:
    s (float): target expected standard deviation
    f (func): a function that takes a list and returns a number
    npts (tuple(int)): a tuple of dimensions of the target product measure
    bounds (tuple, default=None): tuple is ``(lower_bounds, upper_bounds)``
    weights (list, default=None):  a list of sample weights
    tol (float, default=None): maximum allowable deviation from ``s``
    constraints (func, default=None): a function that takes a nested list of
        ``N x 1D`` discrete measure positions and weights, with the intended
        purpose of kernel-transforming ``x,w`` as ``x' = constraints(x, w)``
    npop (int, default=200): size of the trial solution population
    maxiter (int, default=1000): the maximum number of iterations to perform
    maxfun (int, default=1e+6): the maximum number of function evaluations

Returns:
    a list of sample positions, with expected standard deviation ``E``

Notes:
    Expected std ``E`` is calculated by minimizing ``std(f(x)) - s``,
    over the given *bounds*, and will terminate when ``E`` is found within
    deviation ``tol`` of the target std ``s``. If ``tol`` is not provided,
    then a relative deviation of 1% of ``s`` will be used.

    This function does not preserve the mean, variance, or range, as there
    is no initial list of samples to draw the mean, variance, and etc from

    *bounds* is tuple with ``length(bounds) == 2``, composed of all the lower
    bounds, then all the upper bounds, for each parameter

Examples:
    >>> # provide the dimensions and bounds
    >>> nx = 3;  ny = 2;  nz = 1
    >>> x_lb = [10.0];  y_lb = [0.0];  z_lb = [10.0]
    >>> x_ub = [50.0];  y_ub = [9.0];  z_ub = [90.0]
    >>> 
    >>> # prepare the bounds
    >>> lb = (nx * x_lb) + (ny * y_lb) + (nz * z_lb)
    >>> ub = (nx * x_ub) + (ny * y_ub) + (nz * z_ub)
    >>>
    >>> # generate a list of samples with std +/- dev imposed
    >>> std = 2.0;  dev = 0.01
    >>> samples = impose_expected_std(std, f, (nx,ny,nz), (lb,ub), tol=dev)
    >>>
    >>> # test the results by calculating the expected std for the samples
    >>> expected_std(f, samples)
    >>> 2.000010010122465
"""
  # param[0] is the target standard deviation
  # param[1] is the acceptable deviation from the target standard deviation
  tol = kwds['tol'] if 'tol' in kwds else None
  param = (s, tol or 0.01*s)

  # FIXME: the following is a HACK to recover from lost 'weights' information
  #        we 'mimic' discrete measures using the product measure weights
  # plug in the 'constraints' function:  samples' = constrain(samples, weights)
  constrain = None   # default is no constraints
  if 'constraints' in kwds: constrain = kwds['constraints']
  if not constrain:  # if None (default), there are no constraints
    constraints = lambda x: x
  else: #XXX: better to use a standard "xk' = constrain(xk)" interface ?
    def constraints(rv):
      coords = _pack( _nested(rv,npts) )
      coords = list(zip(*coords))              # 'mimic' a nested list
      coords = constrain(coords, [weights for i in range(len(coords))])
      coords = list(zip(*coords))              # revert back to a packed list
      return _flat( _unpack(coords,npts) )

  # construct cost function to reduce deviation from expected std
  def cost(rv):
    """compute cost from a 1-d array of model parameters,
    where:  cost = | E_std[model] - m |**2 """
    # from mystic.math.measures import _pack, _nested, expected_std
    samples = _pack( _nested(rv,npts) )
    Ex = expected_std(f, samples, weights)
    return (Ex - param[0])**2

  # if bounds are not set, use the default optimizer bounds
  if not bounds:
    lower_bounds = []; upper_bounds = []
    for n in npts:
      lower_bounds += [None]*n
      upper_bounds += [None]*n
  else:
    lower_bounds, upper_bounds = bounds

  if len(lower_bounds) == 1 and not (param[1] >= -param[0] >= -param[1]):
     msg = 'if size == 1, then std == 0 (std %s +/- %s)' % param
     raise ValueError(msg)

  # construct and configure optimizer
  debug = kwds['debug'] if 'debug' in kwds else False
  npop = kwds.pop('npop', 200)
  maxiter = kwds.pop('maxiter', 1000)
  maxfun = kwds.pop('maxfun', 1e+6)
  crossover = 0.9; percent_change = 0.9

  def optimize(cost, bounds, tolerance, _constraints):
    (lb,ub) = bounds
    from mystic.solvers import DifferentialEvolutionSolver2
    from mystic.termination import VTR
    from mystic.strategy import Best1Exp
    from mystic.monitors import VerboseMonitor, Monitor
    from mystic.tools import random_seed
    if debug: random_seed(123)
    evalmon = Monitor();  stepmon = Monitor()
    if debug: stepmon = VerboseMonitor(10)

    ndim = len(lb)
    solver = DifferentialEvolutionSolver2(ndim,npop)
    from mystic.termination import VTR
    from mystic.strategy import Best1Exp
    from mystic.monitors import VerboseMonitor, Monitor
    from mystic.tools import random_seed
    if debug: random_seed(123)
    evalmon = Monitor();  stepmon = Monitor()
    if debug: stepmon = VerboseMonitor(10)

    ndim = len(lb)
    solver = DifferentialEvolutionSolver2(ndim,npop)
    solver.SetRandomInitialPoints(min=lb,max=ub)
    solver.SetStrictRanges(min=lb,max=ub)
    solver.SetEvaluationLimits(maxiter,maxfun)
    solver.SetEvaluationMonitor(evalmon)
    solver.SetGenerationMonitor(stepmon)
    solver.Solve(cost,termination=VTR(tolerance),strategy=Best1Exp, \
                 CrossProbability=crossover,ScalingFactor=percent_change, \
                 constraints = _constraints) #XXX: parallel map?

    solved = solver.Solution()
    diameter_squared = solver.bestEnergy
    func_evals = len(evalmon)
    return solved, diameter_squared, func_evals

  # use optimization to get expected std
  tolerance = (param[1])**2
  results = optimize(cost, (lower_bounds, upper_bounds), tolerance, constraints)

  # repack the results
  samples = _pack( _nested(results[0],npts) )
  return samples


#XXX: better ...variance(param, f, ...) or ...variance(m,v, f, ...) ?
def impose_expected_mean_and_variance(param, f, npts, bounds=None, weights=None, **kwds):
  """impose a given expected mean ``E`` on a given function *f*,
where ``E = m +/- tol`` and ``E = mean(f(x))`` for ``x`` in *bounds*.
Additionally, impose a given expected variance ``R`` on *f*,
where ``R = v +/- tol`` and ``R = variance(f(x))`` for ``x`` in *bounds*.

Args:
    param (tuple(float)): target parameters, ``(mean, variance)``
    f (func): a function that takes a list and returns a number
    npts (tuple(int)): a tuple of dimensions of the target product measure
    bounds (tuple, default=None): tuple is ``(lower_bounds, upper_bounds)``
    weights (list, default=None):  a list of sample weights
    tol (float, default=None): maximum allowable deviation from ``m`` and ``v``
    constraints (func, default=None): a function that takes a nested list of
        ``N x 1D`` discrete measure positions and weights, with the intended
        purpose of kernel-transforming ``x,w`` as ``x' = constraints(x, w)``
    npop (int, default=200): size of the trial solution population
    maxiter (int, default=1000): the maximum number of iterations to perform
    maxfun (int, default=1e+6): the maximum number of function evaluations

Returns:
    a list of sample positions, with expected mean ``E`` and variance ``R``

Notes:
    Expected mean ``E`` and expected variance ``R`` are calculated by
    minimizing the sum of the absolute values of ``mean(f(x)) - m`` and
    ``variance(f(x)) - v`` over the given *bounds*, and will terminate when
    ``E`` and ``R`` are found within tolerance ``tol`` of the target mean ``m``
    and variance ``v``, respectively. If ``tol`` is not provided, then a
    relative deviation of 1% of ``max(m,v)`` will be used.

    This function does not preserve the mean, variance, or range, as there
    is no initial list of samples to draw the mean, variance, and etc from

    *bounds* is tuple with ``length(bounds) == 2``, composed of all the lower
    bounds, then all the upper bounds, for each parameter

Examples:
    >>> # provide the dimensions and bounds
    >>> nx = 3;  ny = 2;  nz = 1
    >>> x_lb = [10.0];  y_lb = [0.0];  z_lb = [10.0]
    >>> x_ub = [50.0];  y_ub = [9.0];  z_ub = [90.0]
    >>> 
    >>> # prepare the bounds
    >>> lb = (nx * x_lb) + (ny * y_lb) + (nz * z_lb)
    >>> ub = (nx * x_ub) + (ny * y_ub) + (nz * z_ub)
    >>>
    >>> # generate a list of samples with mean and variance imposed
    >>> mean = 5.0;  var = 2.0;  tol = 0.01
    >>> samples = impose_expected_mean_and_variance((mean,var), f, (nx,ny,nz), \
    ...                                             (lb,ub), tol=tol)
    >>>
    >>> # test the results by calculating the expected mean for the samples
    >>> expected_mean(f, samples)
    >>> 
    >>> # test the results by calculating the expected variance for the samples
    >>> expected_variance(f, samples)
    >>> 2.000010010122465
"""
  # param[0] is the target mean
  # param[1] is the target variance
  # param[2] is the acceptable deviation from the target mean and variance
  tol = kwds['tol'] if 'tol' in kwds else None
  param = tuple(param) + (tol or 0.01*max(param),)

  # FIXME: the following is a HACK to recover from lost 'weights' information
  #        we 'mimic' discrete measures using the product measure weights
  # plug in the 'constraints' function:  samples' = constrain(samples, weights)
  constrain = None   # default is no constraints
  if 'constraints' in kwds: constrain = kwds['constraints']
  if not constrain:  # if None (default), there are no constraints
    constraints = lambda x: x
  else: #XXX: better to use a standard "xk' = constrain(xk)" interface ?
    def constraints(rv):
      coords = _pack( _nested(rv,npts) )
      coords = list(zip(*coords))              # 'mimic' a nested list
      coords = constrain(coords, [weights for i in range(len(coords))])
      coords = list(zip(*coords))              # revert back to a packed list
      return _flat( _unpack(coords,npts) )

  # construct cost function to reduce deviation from expected mean and variance
  def cost(rv):
    """compute cost from a 1-d array of model parameters,
    where:  cost = | E[model] - m |**2 + | E_var[model] - n |**2 """
    # from mystic.math.measures import _pack, _nested
    # from mystic.math.measures import expectation, expected_variance
    samples = _pack( _nested(rv,npts) )
    Em = expectation(f, samples, weights)
    Ev = expected_variance(f, samples, weights)
    return (Em - param[0])**2 + (Ev - param[1])**2

  # if bounds are not set, use the default optimizer bounds
  if not bounds:
    lower_bounds = []; upper_bounds = []
    for n in npts:
      lower_bounds += [None]*n
      upper_bounds += [None]*n
  else:
    lower_bounds, upper_bounds = bounds

  if len(lower_bounds) == 1 and not (param[2] >= -param[1] >= -param[2]):
     msg = 'if size == 1, then variance == 0 (variance %s +/- %s)' % param[1:]
     raise ValueError(msg)

  # construct and configure optimizer
  debug = kwds['debug'] if 'debug' in kwds else False
  npop = kwds.pop('npop', 200)
  maxiter = kwds.pop('maxiter', 1000)
  maxfun = kwds.pop('maxfun', 1e+6)
  crossover = 0.9; percent_change = 0.9

  def optimize(cost, bounds, tolerance, _constraints):
    (lb,ub) = bounds
    from mystic.solvers import DifferentialEvolutionSolver2
    from mystic.termination import VTR
    from mystic.strategy import Best1Exp
    from mystic.monitors import VerboseMonitor, Monitor
    from mystic.tools import random_seed
    if debug: random_seed(123)
    evalmon = Monitor();  stepmon = Monitor()
    if debug: stepmon = VerboseMonitor(10)

    ndim = len(lb)
    solver = DifferentialEvolutionSolver2(ndim,npop)
    solver.SetRandomInitialPoints(min=lb,max=ub)
    solver.SetStrictRanges(min=lb,max=ub)
    solver.SetEvaluationLimits(maxiter,maxfun)
    solver.SetEvaluationMonitor(evalmon)
    solver.SetGenerationMonitor(stepmon)
    solver.Solve(cost,termination=VTR(tolerance),strategy=Best1Exp, \
                 CrossProbability=crossover,ScalingFactor=percent_change, \
                 constraints = _constraints) #XXX: parallel map?

    solved = solver.Solution()
    diameter_squared = solver.bestEnergy
    func_evals = len(evalmon)
    return solved, diameter_squared, func_evals

  # use optimization to get expected mean and variance
  tolerance = (param[-1])**2 #XXX: correct? or (?*D)**2
  results = optimize(cost, (lower_bounds, upper_bounds), tolerance, constraints)

  # repack the results
  samples = _pack( _nested(results[0],npts) )
  return samples


##### weight shift methods #####
def impose_weight_norm(samples, weights, mass=1.0):
  """normalize the weights for a list of (weighted) points
  (this function is 'mean-preserving')

Inputs:
    samples -- a list of sample points
    weights -- a list of sample weights
    mass -- float target of normalized weights
"""
  m = mean(samples, weights)
  wts = normalize(weights,mass) #NOTE: not mean-preserving, until next line
  return impose_mean(m, samples, wts), wts


def normalize(weights, mass='l2', zsum=False, zmass=1.0):
  """normalize a list of points (e.g. normalize to 1.0)

Inputs:
    weights -- a list of sample weights
    mass -- float target of normalized weights (or string for Ln norm)
    zsum -- use counterbalance when mass = 0.0
    zmass -- member scaling when mass = 0.0

Notes: if mass='l1', will use L1-norm; if mass='l2' will use L2-norm; etc.
"""
  try:
    mass = int(mass.lstrip('l'))
    fixed = False
  except AttributeError:
    fixed = True
  weights = asarray(list(weights)) #XXX: faster to use x = array(x, copy=True) ?

  if fixed:
    w = sum(abs(weights))
  else:
    mass = int(min(200, mass)) # x**200 is ~ x**inf
    w = Lnorm(weights,mass)
    mass = 1.0

  if not w:
    if not zsum: return list(weights * 0.0)
    from numpy import inf, nan
    weights[weights == 0.0] = nan
    return list(weights * inf)  # protect against ZeroDivision

  if float(mass) or not zsum:
    w = weights / w #FIXME: not "mean-preserving"
    if not fixed: return list(w) # <- scaled so sum(abs(x)) = 1
    #REMAINING ARE fixed mean
    m = sum(w)
    w = mass * w
    if not m:  #XXX: do similar to zsum (i.e. shift) when sum(weights)==0 ?
      if not zsum: return list(weights * 0.0)
      from numpy import inf, nan
      weights[weights == 0.0] = nan
      return list(weights * inf)  # protect against ZeroDivision
    return list(w/m) # <- scaled so sum(x) = 1

  # force selected member to satisfy sum = 0.0
  zsum = -1
  weights[zsum] = -(sum(weights) - weights[zsum])
  mass = zmass
  return list(mass * weights / w)  #FIXME: not "mean-preserving"


def impose_reweighted_mean(m, samples, weights=None, solver=None):
    """impose a mean on a list of points by reweighting weights"""
    ndim = len(samples)
    if weights is None:
        weights = [1.0/ndim] * ndim
    if solver is None or solver == 'fmin':
        from mystic.solvers import fmin as solver
    elif solver == 'fmin_powell':
        from mystic.solvers import fmin_powell as solver
    elif solver == 'diffev':
        from mystic.solvers import diffev as solver
    elif solver == 'diffev2':
        from mystic.solvers import diffev2 as solver
    norm = sum(weights)

    inequality = ""; equality = ""; equality2 = ""
    for i in range(ndim):
        inequality += "x%s >= 0.0\n" % (i) # positive
        equality += "x%s + " % (i)         # normalized
        equality2 += "%s * x%s + " % (float(samples[i]),(i)) # mean

    equality += "0.0 = %s\n" % float(norm)
    equality += equality2 + "0.0 = %s*%s\n" % (float(norm),m)

    penalties = generate_penalty(generate_conditions(inequality))
    constrain = generate_constraint(generate_solvers(solve(equality)))

    def cost(x): return sum(x)

    results = solver(cost, weights, constraints=constrain, \
                     penalty=penalties, disp=False, full_output=True)
    wts = list(results[0])
    _norm = results[1] # should have _norm == norm
    warn = results[4]  # nonzero if didn't converge

    #XXX: better to fail immediately if xlo < m < xhi... or the below?
    if warn or not almostEqual(_norm, norm):
        print("Warning: could not impose mean through reweighting")
        return None #impose_mean(m, samples, weights), weights

    return wts #samples, wts


def impose_reweighted_variance(v, samples, weights=None, solver=None):
    """impose a variance on a list of points by reweighting weights"""
    ndim = len(samples)
    if weights is None:
        weights = [1.0/ndim] * ndim
    if solver is None or solver == 'fmin':
        from mystic.solvers import fmin as solver
    elif solver == 'fmin_powell':
        from mystic.solvers import fmin_powell as solver
    elif solver == 'diffev':
        from mystic.solvers import diffev as solver
    elif solver == 'diffev2':
        from mystic.solvers import diffev2 as solver
    norm = sum(weights)
    m = mean(samples, weights)

    inequality = ""
    equality = ""; equality2 = ""; equality3 = ""
    for i in range(ndim):
        inequality += "x%s >= 0.0\n" % (i) # positive
        equality += "x%s + " % (i)         # normalized
        equality2 += "%s * x%s + " % (float(samples[i]),(i)) # mean
        equality3 += "x%s*(%s-%s)**2 + " % ((i),float(samples[i]),m) # var

    equality += "0.0 = %s\n" % float(norm)
    equality += equality2 + "0.0 = %s*%s\n" % (float(norm),m)
    equality += equality3 + "0.0 = %s*%s\n" % (float(norm),v)

    penalties = generate_penalty(generate_conditions(inequality))
    constrain = generate_constraint(generate_solvers(solve(equality)))

    def cost(x): return sum(x)

    results = solver(cost, weights, constraints=constrain, \
                     penalty=penalties, disp=False, full_output=True)
    wts = list(results[0])
    _norm = results[1] # should have _norm == norm
    warn = results[4]  # nonzero if didn't converge

    #XXX: better to fail immediately if xlo < m < xhi... or the below?
    if warn or not almostEqual(_norm, norm):
        print("Warning: could not impose mean through reweighting")
        return None #impose_variance(v, samples, weights), weights

    return wts #samples, wts  # "mean-preserving"

def impose_reweighted_std(s, samples, weights=None, solver=None):
    """impose a standard deviation on a list of points by reweighting weights"""
    return impose_reweighted_variance(s**2, samples, weights, solver)

##### sampling statistics methods #####
def _sort(samples, weights=None):
    "sort (weighted) samples; returns 2-D array of samples and weights"
    import numpy as np
    if weights is None:
        x = np.ones((2,len(samples))) #XXX: casts to a float
        x[0] = np.sort(samples)
        return x
    x = np.vstack([samples,weights]).T
    return x[x[:,0].argsort()].T


def median(samples, weights=None):
    """calculate the (weighted) median for a list of points

Inputs:
    samples -- a list of sample points
    weights -- a list of sample weights
"""
    import numpy as np
    x,w = _sort(samples,weights)
    s = sum(w)
    return np.mean(x[s/2. - np.cumsum(w) <= 0][0:2-x.size%2])


def mad(samples, weights=None): #, scale=1.4826):
    """calculate the (weighted) median absolute deviation for a list of points

Inputs:
    samples -- a list of sample points
    weights -- a list of sample weights
"""
    s = asarray(samples)
    return median(abs(s - median(samples,weights)),weights) # * scale


#XXX: impose median by ensuring equal # samples < & >... (and not shifting) ?
def impose_median(m, samples, weights=None):
    """impose a median on a list of (weighted) points
    (this function is 'range-preserving' and 'mad-preserving')

Inputs:
    m -- the target median
    samples -- a list of sample points
    weights -- a list of sample weights
"""
    s = asarray(samples)
    return (s + (m - median(samples, weights))).tolist()


def impose_mad(s, samples, weights=None):
    """impose a median absolute deviation on a list of (weighted) points
    (this function is 'median-preserving')

Inputs:
    s -- the target median absolute deviation
    samples -- a list of sample points
    weights -- a list of sample weights
"""
    import numpy as np
    m = median(samples, weights)
    samples = np.asarray(list(samples))
    _mad = mad(samples,weights)
    if not _mad: # protect against ZeroDivision when mad = 0
        return [np.nan]*len(samples)
    scale = float(s) / _mad
    samples = samples * scale #NOTE: not "median-preserving" until next line
    return impose_median(m, samples, weights) #NOTE: not "range-preserving"


def _k(weights, k=0, clip=False, norm=False, eps=15): #XXX: better 9 ?
    "trim weights at k%; if clip is True, winsorize instead of trim"
    #NOTE: eps is tolerance for cancellation of similar values
    import numpy as np
    try:
        klo,khi = k
    except TypeError:
        klo = khi = k
    if klo + khi > 100:
        msg = "cannot crop '%s + %s > 100' percent" % (klo,khi)
        raise ValueError(msg)
    elif klo < 0 or khi < 0:
        msg = "cannot crop negative percent '%s + %s'" % (klo,khi)
        raise ValueError(msg)
    else:
        klo,khi = .01*klo,.01*khi
    w = np.array(weights, dtype=float)/sum(weights)  #XXX: no dtype?
    w_lo, w_hi = np.cumsum(w), np.cumsum(w[::-1])
    # calculate the cropped indices
    lo = len(w) - sum((w_lo - klo).round(eps) > 0)
    hi = sum((w_hi - khi).round(eps) > 0) - 1
    # flip indices if flipped
    if lo > hi:
        lo,hi = hi,lo
    if not clip:
        # find the values at k%
        w_lo = w_lo[lo]
        w_hi = w_hi[len(w)-1-hi]
        # reset the values at k%
        if klo + khi == 1:
            w[lo] = w[hi] = 0
        elif lo == hi:
            w[lo] = max(1 - khi - klo,0)
        else: 
            w[lo] = max(w_lo - klo,0)
            w[hi] = max(w_hi - khi,0)
    else:
        # reset the values at k%
        if lo == hi:
            w[lo] = sum(w)
        else: 
            w[lo] += sum(w[:lo])
            w[hi] += sum(w[hi+1:])
    # trim the remaining weights
    w[:lo] = 0
    w[hi+1:] = 0
    if not norm: w *= sum(weights)
    return w.tolist()


def tmean(samples, weights=None, k=0, clip=False):
    """calculate the (weighted) trimmed mean for a list of points

Inputs:
    samples -- a list of sample points
    weights -- a list of sample weights
    k -- percent samples to trim (k%) [tuple (lo,hi) or float if lo=hi]
    clip -- if True, winsorize instead of trimming k% of samples

NOTE: if all samples are excluded, will return nan
"""
    samples,weights = _sort(samples,weights)
    weights = _k(weights,k,clip)
    return sum(samples * weights)/sum(weights)


def tvariance(samples, weights=None, k=0, clip=False):
    """calculate the (weighted) trimmed variance for a list of points

Inputs:
    samples -- a list of sample points
    weights -- a list of sample weights
    k -- percent samples to trim (k%) [tuple (lo,hi) or float if lo=hi]
    clip -- if True, winsorize instead of trimming k% of samples

NOTE: if all samples are excluded, will return nan
"""
    samples,weights = _sort(samples,weights)
    weights = _k(weights,k,clip)
    trim_mean = sum(samples * weights)/sum(weights)
    return mean(abs(samples - trim_mean)**2, weights) #XXX: correct ?


def tstd(samples, weights=None, k=0, clip=False):
    """calculate the (weighted) trimmed standard deviation for a list of points

Inputs:
    samples -- a list of sample points
    weights -- a list of sample weights
    k -- percent samples to trim (k%) [tuple (lo,hi) or float if lo=hi]
    clip -- if True, winsorize instead of trimming k% of samples

NOTE: if all samples are excluded, will return nan
"""
    import numpy as np
    return np.sqrt(tvariance(samples, weights, k, clip))


#XXX: use reweighting to impose tmean, tvariance, tstd, median, & mad ?
def impose_tmean(m, samples, weights=None, k=0, clip=False):
    """impose a trimmed mean (at k%) on a list of (weighted) points
    (this function is 'range-preserving' and 'tvariance-preserving')

Inputs:
    m -- the target trimmed mean
    samples -- a list of sample points
    weights -- a list of sample weights
    k -- percent samples to be trimmed (k%) [tuple (lo,hi) or float if lo=hi]
    clip -- if True, winsorize instead of trimming k% of samples
"""
    s = asarray(samples)
    return (s + (m - tmean(samples, weights, k=k, clip=clip))).tolist()


def impose_tvariance(v, samples, weights=None, k=0, clip=False):
    """impose a trimmed variance (at k%) on a list of (weighted) points
    (this function is 'tmean-preserving')

Inputs:
    v -- the target trimmed variance
    samples -- a list of sample points
    weights -- a list of sample weights
    k -- percent samples to be trimmed (k%) [tuple (lo,hi) or float if lo=hi]
    clip -- if True, winsorize instead of trimming k% of samples
"""
    import numpy as np
    m = tmean(samples, weights, k=k, clip=clip)
    samples = np.asarray(list(samples))

    tvar = tvariance(samples,weights,k=k,clip=clip)
    if not tvar: # protect against ZeroDivision when tvar = 0
        return [np.nan]*len(samples) #XXX: k?
    scale = np.sqrt(float(v) / tvar)
    samples = samples * scale #NOTE: not "tmean-preserving" until next line
    return impose_tmean(m, samples, weights, k=k, clip=clip) #NOTE: not "range-preserving"


def impose_tstd(s, samples, weights=None, k=0, clip=False):
    """impose a trimmed std (at k%) on a list of (weighted) points
    (this function is 'tmean-preserving')

Inputs:
    s -- the target trimmed standard deviation
    samples -- a list of sample points
    weights -- a list of sample weights
    k -- percent samples to be trimmed (k%) [tuple (lo,hi) or float if lo=hi]
    clip -- if True, winsorize instead of trimming k% of samples
"""
    return impose_tvariance(s**2, samples, weights, k=k, clip=clip)


##### collapse methods #####
#FIXME: add collapse weight/position methods to math.discrete.measure?
def impose_support(index, samples, weights): #XXX: toggle norm_preserving?
    """set all weights not appearing in 'index' to zero

Inputs:
    samples -- a list of sample points
    weights -- a list of sample weights
    index -- a list of desired support indices (weights will be non-zero)

For example:
    >>> impose_support([0,1],[1,2,3,4,5],[.2,.2,.2,.2,.2])
    ([2.5, 3.5, 4.5, 5.5, 6.5], [0.5, 0.5, 0.0, 0.0, 0.0])
    >>> impose_support([0,1,2,3],[1,2,3,4,5],[.2,.2,.2,.2,.2])
    ([1.5, 2.5, 3.5, 4.5, 5.5], [0.25, 0.25, 0.25, 0.25, 0.0])
    >>> impose_support([4],[1,2,3,4,5],[.2,.2,.2,.2,.2])
    ([-1.0, 0.0, 1.0, 2.0, 3.0], [0.0, 0.0, 0.0, 0.0, 1.0])

Notes: is 'mean-preserving' for samples and 'norm-preserving' for weights
"""
    if index is None: index = range(len(weights))
    # allow negative indexing
    index = set(len(weights)+i if i<0 else i for i in index)
    m = mean(samples, weights)
    n = sum(weights)
    weights = [w if i in index else 0. for (i,w) in enumerate(weights)]
    weights = normalize(weights, n)
    return impose_mean(m, samples, weights), weights


#XXX: alternate to the above
def impose_unweighted(index, samples, weights, nullable=True):
    """set all weights appearing in 'index' to zero

Inputs:
    samples -- a list of sample points
    weights -- a list of sample weights
    index -- a list of indices where weight is to be zero
    nullable -- if False, avoid null weights by reweighting non-index weights

For example:
    >>> impose_unweighted([0,1,2],[1,2,3,4,5],[.2,.2,.2,.2,.2])
    ([-0.5, 0.5, 1.5, 2.5, 3.5], [0.0, 0.0, 0.0, 0.5, 0.5])
    >>> impose_unweighted([3,4],[1,2,3,4,5],[.2,.2,.2,.2,.2])
    ([2.0, 3.0, 4.0, 5.0, 6.0], [0.33333333333333331, 0.33333333333333331, 0.33333333333333331, 0.0, 0.0])

Notes: is 'mean-preserving' for samples and 'norm-preserving' for weights
"""
    if index is None: index = ()
    # allow negative indexing
    index = set(len(weights)+i if i<0 else i for i in index)
    m = mean(samples, weights)
    n = sum(weights)
    _weights = [0. if i in index else w for (i,w) in enumerate(weights)]
    if not nullable and not sum(_weights):
        _weights = [0. if i in index else 1. for (i,w) in enumerate(weights)]
    weights = normalize(_weights, n)
    return impose_mean(m, samples, weights), weights


def impose_collapse(pairs, samples, weights):
    """collapse the weight and position of each pair (i,j) in pairs

Collapse is defined as weight[j] += weight[i] and weights[i] = 0,
with samples[j] = samples[i].

Inputs:
    samples -- a list of sample points
    weights -- a list of sample weights
    pairs -- set of tuples of indices (i,j) where collapse occurs

For example:
    >>> impose_collapse({(0,1),(0,2)},[1,2,3,4,5],[.2,.2,.2,.2,.2])
    ([1.5999999999999996, 1.5999999999999996, 1.5999999999999996, 4.5999999999999996, 5.5999999999999996], [0.6000000000000001, 0.0, 0.0, 0.2, 0.2])
    >>> impose_collapse({(0,1),(3,4)},[1,2,3,4,5],[.2,.2,.2,.2,.2])
    ([1.3999999999999999, 1.3999999999999999, 3.3999999999999999, 4.4000000000000004, 4.4000000000000004], [0.4, 0.0, 0.2, 0.4, 0.0])

Notes: is 'mean-preserving' for samples and 'norm-preserving' for weights
"""
    samples, weights = list(samples), list(weights) # don't edit inputs
    m = mean(samples, weights)
    # allow negative indexing
    pairs = zip(*tuple(tuple(len(weights)+i if i<0 else i for i in j) for j in zip(*pairs)))
    #XXX: any vectorized way to do this?
    from mystic.tools import connected
    pairs = connected(pairs)
    for i,j in getattr(pairs,'iteritems',pairs.items)():
        v = weights[i]
        for k in j:
            v += weights[k]
            weights[k] = type(v)(0.0)
            samples[k] = samples[i]
        weights[i] = v
    return impose_mean(m, samples, weights), weights
    #XXX: any vectorized way to do this?
#   for i,j in sorted(sorted(pair) for pair in pairs): 
#       weight = weights[i]+weights[j]
#       weights[i],weights[j] = type(weight)(0.0),weight #FIXME: also wrong
#       samples[j] = samples[i] #FIXME: all paired collapses should be equal
#   return impose_mean(m, samples, weights), weights


##### misc methods #####
def impose_sum(mass, weights, zsum=False, zmass=1.0):
  """impose a sum on a list of points

Inputs:
    mass -- target sum of weights
    weights -- a list of sample weights
    zsum -- use counterbalance when mass = 0.0
    zmass -- member scaling when mass = 0.0
"""
  return normalize(weights, mass, zsum, zmass)

def impose_product(mass, weights, zsum=False, zmass=1.0):
  """impose a product on a list of points

Inputs:
    mass -- target product of weights
    weights -- a list of sample weights
    zsum -- use counterbalance when mass = 0.0
    zmass -- member scaling when mass = 0.0
"""
  from numpy import product
  weights = asarray(list(weights)) #XXX: faster to use x = array(x, copy=True) ?
  w = float(product(weights))
  n = len(weights)
  if not w:  #XXX: is this the best behavior?
    from numpy import inf
    return list(weights * inf)  # protect against ZeroDivision
  if float(mass):
    if w/mass < 0.0:
      return list(-weights / (-w/mass)**(1./n))  #FIXME: not "mean-preserving"
    return list(weights / (w/mass)**(1./n))      #FIXME: not "mean-preserving"
  # force selected member to satisfy product = 0.0
  if not zsum:
    return list(weights * 0.0)  #FIXME: not "mean-preserving"
  zsum = -1
  p, weights[zsum] = weights[zsum], 0.0
  w = (w/p)
  n = n-1
  mass = zmass
  if w/mass >= 0.0:
    return list(weights[:-1]/(w/mass)**(1./n))+[0.]#FIXME: not "mean-preserving"
  return list(-weights[:-1]/(-w/mass)**(1./n))+[0.]#FIXME: not "mean-preserving"


#--------------------------------------------------------------------
# <helper methods>

##### packing / unpacking
# >>> a = [1,2,3]; b = [4,5]
# >>> zip(a[0:]+a[:0],b) + zip(a[1:]+a[:1],b) + zip(a[2:]+a[:2],b)
# [(1, 4), (2, 5), (3, 6), (2, 4), (3, 5), (1, 6), (3, 4), (1, 5), (2, 6)]
def _pack(samples):
  """'pack' a list of discrete measure sample points 
into a list of product measure sample points

Inputs:
    samples -- a list of sample points for N discrete measures

For example:
  >>> _pack([[1,2,3], [4,5], [6,7]])
  [(1,4,6), (2,4,6), (3,4,6), (1,5,6), (2,5,6), (3,5,6), \
   (1,4,7), (2,4,7), (3,4,7), (1,5,7), (2,5,7), (3,5,7)]
"""
 #from numpy import product, array, ones
 #ndim = len(samples)
 #npts = [len(s) for s in samples]
 #z = []
 #for i in range(ndim):
 #  tmp = list(array([n*ones(product(npts[:i])) for n in samples[i]]).flatten())
 #  z.append( product(npts[i+1:])*tmp )
 #del tmp
 #zT = []
 #for i in range(len(z[0])):
 #  zT.append( tuple([y.pop(0) for y in z]) )
 #return zT
# from numpy import product, array, ones
# ndim = len(samples)
# npts = [len(s) for s in samples]
# z = ones((ndim, product(npts)))  # z.T of what's needed
# for i in range(ndim):
#   tmp = list(array([n*ones(product(npts[:i])) for n in samples[i]]).flatten())
#   z[i] = product(npts[i+1:])*tmp
# return [tuple(i) for i in z.T]
  ndim = len(samples)
  currentx=[0.]*ndim
  _samples = []
  def recurse(next):
    if next == -1:
      _samples.append(tuple(currentx))
      return
    else:
      for xpt in samples[next]:
        currentx[next]=xpt
        recurse(next - 1)
  recurse(ndim-1)
  return _samples

def _unpack(samples, npts):
  """'unpack' a list of product measure sample points 
into a list of discrete measure sample points

Inputs:
    samples -- a list of sample points for a product measure
    npts -- a tuple of dimensions of the target discrete measure

For example:
  >>> _unpack( [(1,4,6), (2,4,6), (3,4,6), (1,5,6), (2,5,6), (3,5,6), \
  ...           (1,4,7), (2,4,7), (3,4,7), (1,5,7), (2,5,7), (3,5,7)], (3,2,2) \
  ...        )
  [[1,2,3], [4,5], [6,7]]
"""
# from numpy import product, array
# ndim = len(npts)
# z = []
# for i in range(ndim):
#   tmp = array(samples[:int(len(samples)/product(npts[i+1:]))]).T[i]
#   z.append( list(tmp[::int(product(npts[:i]))]) )
# return z
  _samples = []
  ndim = len(npts)
  temp = [npts[0]]*ndim
  _samples.append([j[0] for j in samples][:npts[0]])
  def recurse(next):
    if next == ndim:
      return
    else:
      temp[next] = temp[next-1]*npts[next]
      currentindex = temp[next]
      lastindex = temp[next-1]
      _samples.append([j[next] for j in samples][:currentindex:lastindex])
      recurse(next + 1)
  recurse(1)
  return _samples


def _flat(params):
  """
converts a nested parameter list into flat parameter list

Inputs:
    params -- a nested list of weights or positions

For example:
    >>> par = [['x','x','x'], ['y','y'], ['z']]
    >>> _flat(par)
    ['x','x','x','y','y','z']
"""
  from mystic.tools import flatten
  return list(flatten(params))


def _nested(params, npts):
  """
converts a flat parameter list into nested parameter list

Inputs:
    params -- a flat list of weights or positions
    npts -- a tuple describing the shape of the target list

For example:
    >>> nx = 3;  ny = 2;  nz = 1
    >>> par = ['x']*nx + ['y']*ny + ['z']*nz
    >>> _nested(par, (nx,ny,nz))
    [['x','x','x'], ['y','y'], ['z']]
"""
  coords = []
  ind = 0
  for i in range(len(npts)):
    coords.append(params[ind:ind+npts[i]])
    ind += npts[i]
  return coords


def _nested_split(params, npts):
  """
splits a flat parameter list into a list of weights and a list of positions;
weights and positions are expected to have the same dimensions (given by npts)

Inputs:
    params -- a flat list of weights and positions (formatted as noted below)
    npts -- a tuple describing the shape of the target lists

For example:
    >>> nx = 3;  ny = 2;  nz = 1
    >>> par = ['wx']*nx + ['x']*nx + ['wy']*ny + ['y']*ny + ['wz']*nz + ['z']*nz
    >>> weights, positions = _nested_split(par, (nx,ny,nz))
    >>> weights
    [['wx','wx','wx'], ['wy','wy'], ['wz']]
    >>> positions
    [['x','x','x'], ['y','y'], ['z']]
"""
  weights = []
  coords = []
  ind = 0
  for i in range(len(npts)):
    weights.append(params[ind:ind+npts[i]])
    ind += npts[i]
    coords.append(params[ind:ind+npts[i]])
    ind += npts[i]
  return weights, coords


def split_param(params, npts):
  """
splits a flat parameter list into a flat list of weights
and a flat list of positions;  weights and positions are expected
to have the same dimensions (given by npts)

Inputs:
    params -- a flat list of weights and positions (formatted as noted below)
    npts -- a tuple describing the shape of the target lists

For example:
    >>> nx = 3;  ny = 2;  nz = 1
    >>> par = ['wx']*nx + ['x']*nx + ['wy']*ny + ['y']*ny + ['wz']*nz + ['z']*nz
    >>> weights, positions = split_param(par, (nx,ny,nz))
    >>> weights
    ['wx','wx','wx','wy','wy','wz']
    >>> positions
    ['x','x','x','y','y','z']
"""
  w,x = _nested_split(params, npts)
  return _flat(w), _flat(x)


# backward compatibility
_flat_split = split_param

# EOF



#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2006-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
'''a signal handler for a mystic solver instance'''

try:
    input = raw_input
except NameError:
    pass

# pull from the signal module
from signal import *

class Handler(object):
    def __init__(self, solver, sigint_callback=None):
        """factory to generate signal handler

Inputs::
    - solver: a mystic.solver instance
    - sigint_callback: a callback function

Available switches::
    - sol  --> Print current best solution.
    - cont --> Continue calculation.
    - call --> Executes sigint_callback, if provided.
    - exit --> Exits with current best solution.
"""
        self.solver = solver
        self.sigint_callback = sigint_callback or solver.sigint_callback
    def __call__(self, signum, frame):
        import inspect
        print(inspect.getframeinfo(frame))
        print(inspect.trace())
        while 1:
            s = input(\
"""
 
 Enter sense switch.

    sol:  Print current best solution.
    cont: Continue calculation.
    call: Executes sigint_callback [%s].
    exit: Exits with current best solution.

 >>> """ % self.sigint_callback)
            if s.lower() == 'sol':
                print(self.solver.bestSolution)
            elif s.lower() == 'cont':
                return
            elif s.lower() == 'call':
                # sigint call_back
                if self.sigint_callback is not None:
                    self.sigint_callback(self.solver.bestSolution)
            elif s.lower() == 'exit':
                self.solver._EARLYEXIT = True
                return
            else:
                print("unknown option : %s" % s)
        return

# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2020-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

import mystic.cache as mc
import mystic.models as mm

# basic interaction with an archive
d = mc.archive.read('test', type=mc.archive.dict_archive)
assert len(d) == 0

mc.archive.write(d, dict(a=1, b=2, c=3))
assert len(d) == 3

# basic pattern to cache an objective
d = mc.archive.read('rosen', type=mc.archive.dict_archive)
model = mc.cached(archive=d)(mm.rosen)
model([1,2,1])
model([1,1,1])
c = model.__cache__()
assert len(c) == 2

model.__inverse__([1,2,3]) == -model([1,2,3])
assert len(c) == 3



#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Given a set of points in the plane, find the smallest circle
that contains them. (using DE and scipy.fmin)

Requires:
  -- numpy, matplotlib

The matplotlib output will draw 
  -- a set of points inside a circle defined by x0,y0,R0 
  -- the circle (x0,y0) with rad R0
  -- the optimized circle with minimum R enclosing the points
"""

from mystic.models import circle, sparse_circle
import matplotlib.pyplot as plt

# generate training set & define cost function
# CostFactory2 allows costfunction to reuse datapoints from training set
x0, y0, R0 = [10., 20., 3]
npts = 20
xy = sparse_circle(x0, y0, R0, npts)
cost = sparse_circle.CostFactory2(xy)

# function to find the 'support vectors' given R
# SV in quotes because they are found by optimizing the primal,
# as opposed to "directly" via the dual.
def sv(data, xx,yy,rr):
    svl = []
    for i in range(len(data)):
       x,y = data[i]
       if abs((xx-x)*(xx-x)+(yy-y)*(yy-y) - rr*rr) < 0.01:
           svl.append(i)
    return svl

# DEsolver inputs
MAX_GENERATIONS = 2000
ND, NP = 3, 30    # dimension, population size
minrange = [0., 0., 0.]
maxrange = [50., 50., 10.]

# prepare DESolver
from mystic.solvers import DifferentialEvolutionSolver2 \
      as DifferentialEvolutionSolver
from mystic.termination import ChangeOverGeneration, VTR
solver = DifferentialEvolutionSolver(ND, NP)
solver.enable_signal_handler()
solver.SetRandomInitialPoints(min=minrange,max=maxrange)
solver.SetEvaluationLimits(generations=MAX_GENERATIONS)
solver.Solve(cost, termination=ChangeOverGeneration(generations=100))


if __name__ == '__main__':     

    # x0, y0, R0
    #guess = [1,1,1] # bad initial guess
    #guess = [5,5,1] # ok guess
    guess = [10,15,5] # good initial guess

    # plot training set & training set boundary
    plt.plot(xy[:,0],xy[:,1],'k+',markersize=6)
    c = circle(x0, y0, R0)
    plt.plot(c[:,0],c[:,1],'r-',linewidth=2)
    legend = ['random points','generating circle : %f' % R0]
    plt.axis('equal')

    # solve with mystic's differential evolution solver
    solution = solver.Solution()
    sx, sy, sr = solution
    print("DEsol : (%f, %f) @ R = %f" % (sx, sy, sr))

    # plot DEsolver solution
    c = circle(sx, sy, sr)
    plt.plot(c[:,0],c[:,1],'b-',linewidth=2)
    legend.append('DE optimal : %f' % sr)

    # solve with scipy.fmin
    from mystic.solvers import fmin
    sol = fmin(cost, guess)
    print("scipy.fmin sol: %s" % sol)
    ax, ay, ar = sol

    # plot scipy.fmin solution
    c = circle(ax, ay, ar)
    plt.plot(c[:,0],c[:,1],'g-',linewidth=2)
    legend.append('Nelder-Mead : %f' % ar)

    # solve with scipy.brute
   #from mystic._scipyoptimize import brute
   #ranges = tuple(zip(minrange,maxrange))
   #sol = brute(cost, ranges, Ns=NP)
   #print("scipy.brute sol: %s" % sol)
   #bx, by, br = sol

    # plot scipy.brute solution
   #c = circle(bx, by, br)
   #plt.plot(c[:,0],c[:,1],'y-',linewidth=2)
   #legend.append('Brute : %f' % br)

    # find & draw the support vectors from DE
    svl = sv(xy, sx,sy,sr)
    print("DE support vectors: %s" % svl)
    plt.plot(xy[svl,0],xy[svl,1],'bx',markersize=6)

    # find & draw the support vectors from scipy.brute
   #svl = sv(xy, bx,by,br)
   #print("Brute support vectors: %s" % svl)
   #plt.plot(xy[svl,0],xy[svl,1],'yx',markersize=6)

    plt.legend(legend)
    plt.show()

# $Id$
# 
# end of file



#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
test_getCost.py
Example to demonstrate use of CostFactory
"""

from mystic.termination import *
from mystic.strategy import *
from forward_model import *

from mystic.math import poly1d as ForwardPolyFactory
from mystic.models import poly; PolyCostFactory = poly.CostFactory
from mystic.solvers import DifferentialEvolutionSolver
from mystic.monitors import VerboseMonitor
from mystic.tools import getch

ND = 3
NP = 80
MAX_GENERATIONS = ND*NP

from numpy import array

def data(params):
    fwd = ForwardPolyFactory(params)
    x = (array([list(range(101))])-50.)[0]
    return x,fwd(x)

def de_solve(CF):
    solver = DifferentialEvolutionSolver(ND, NP)
    solver.enable_signal_handler()

    stepmon = VerboseMonitor(10,50)
    minrange = [-100., -100., -100.];
    maxrange = [100., 100., 100.];
    solver.SetRandomInitialPoints(min = minrange, max = maxrange)
    solver.SetStrictRanges(min = minrange, max = maxrange)
    solver.SetEvaluationLimits(generations=MAX_GENERATIONS)
    solver.SetGenerationMonitor(stepmon)

    solver.Solve(CF, termination=ChangeOverGeneration(generations=300),\
                CrossProbability=0.5, ScalingFactor=0.5,\
                sigint_callback=plot_sol)

    solution = solver.Solution()
    return solution, stepmon

def plot_sol(params,linestyle='b-'):
    d = data(params)
    plt.plot(d[0],d[1],'%s'%linestyle,linewidth=2.0)
    plt.axis(plotview)
    plt.draw()
    plt.pause(0.001)
    return

from numpy import sum as numpysum
def cost_function(params):
    x = data(params)[1] - datapts
    return numpysum(real((conjugate(x)*x)))


if __name__ == '__main__':
    plotview = [-60,60, 0,2500]
    target = [1., 2., 1.]
    x,datapts = data(target)

   #myCost = cost_function
  ##myCost = PolyCostFactory(target,x)
    F = CostFactory()
    F.addModel(ForwardPolyFactory,len(target),'poly')
    myCost = F.getCostFunction(evalpts=x, observations=datapts)    

    import matplotlib.pyplot as plt
    plt.ion()

    print("target: %s" % target)
    plot_sol(target,'r-')
    solution, stepmon = de_solve(myCost)
    print("solution: %s" % solution)
    plot_sol(solution,'g-')
    print("")
#   print("at step 10: %s" % stepmon.x[10])

    getch()

# End of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Example:
    - Solve 8th-order Chebyshev polynomial coefficients with Nelder-Mead.
    - Callable plot of fitting to Chebyshev polynomial.
    - Plot (x2) of convergence to Chebyshev polynomial.
    - Monitor (x2) Chi-Squared for Chebyshev polynomial.

Demonstrates:
    - standard models
    - expanded solver interface
    - parameter bounds constraints
    - solver interactivity
    - customized monitors and termination conditions
"""

# Nelder-Mead Simplex solver
from mystic.solvers import NelderMeadSimplexSolver

# Chebyshev polynomial and cost function
from mystic.models.poly import chebyshev8, chebyshev8cost
from mystic.models.poly import chebyshev8coeffs

# tools
from mystic.termination import CandidateRelativeTolerance as CRT
from mystic.monitors import VerboseMonitor, Monitor
from mystic.tools import getch
from mystic.math import poly1d
import matplotlib.pyplot as plt
plt.ion()

# draw the plot
def plot_frame(label=None):
    plt.close()
    plt.title("8th-order Chebyshev coefficient convergence")
    plt.xlabel("Nelder-Mead Simplex Solver %s" % label)
    plt.ylabel("Chi-Squared")
    plt.draw()
    plt.pause(0.001)
    return
 
# plot the polynomial trajectories
def plot_params(monitor):
    x = list(range(len(monitor)))
    y = monitor.y
    plt.plot(x,y,'b-')
    plt.axis([1,0.5*x[-1],0,y[1]])#,'k-')
    plt.draw()
    plt.pause(0.001)
    return

# draw the plot
def plot_exact():
    plt.title("fitting 8th-order Chebyshev polynomial coefficients")
    plt.xlabel("x")
    plt.ylabel("f(x)")
    import numpy
    x = numpy.arange(-1.2, 1.2001, 0.01)
    exact = chebyshev8(x)
    plt.plot(x,exact,'b-')
    plt.legend(["Exact"])
    plt.axis([-1.4,1.4,-2,8])#,'k-')
    plt.draw()
    plt.pause(0.001)
    return
 
# plot the polynomial
def plot_solution(params,style='y-'):
    import numpy
    x = numpy.arange(-1.2, 1.2001, 0.01)
    f = poly1d(params)
    y = f(x)
    plt.plot(x,y,style)
    plt.legend(["Exact","Fitted"])
    plt.axis([-1.4,1.4,-2,8])#,'k-')
    plt.draw()
    plt.pause(0.001)
    return

if __name__ == '__main__':

    print("Nelder-Mead Simplex")
    print("===================")

    # initial guess
    import random
    from mystic.tools import random_seed
    random_seed(123)
    ndim = 9
    x0 = [random.uniform(-5,5) + chebyshev8coeffs[i] for i in range(ndim)]

    # suggest that the user interacts with the solver
    print("NOTE: while solver is running, press 'Ctrl-C' in console window")
    getch()

    # draw frame and exact coefficients
    plot_exact()

    # select parameter bounds constraints
    from numpy import inf
    min_bounds = [  0,-1,-300,-1,  0,-1,-100,-inf,-inf]
    max_bounds = [200, 1,   0, 1,200, 1,   0, inf, inf]

    # configure monitors
    stepmon = VerboseMonitor(100)
    evalmon = Monitor()

    # use Nelder-Mead to solve 8th-order Chebyshev coefficients
    solver = NelderMeadSimplexSolver(ndim)
    solver.SetInitialPoints(x0)
    solver.SetEvaluationLimits(generations=999)
    solver.SetEvaluationMonitor(evalmon)
    solver.SetGenerationMonitor(stepmon)
    solver.SetStrictRanges(min_bounds,max_bounds)
    solver.enable_signal_handler()
    solver.Solve(chebyshev8cost, termination=CRT(1e-4,1e-4), \
                 sigint_callback=plot_solution)
    solution = solver.bestSolution

    # get solved coefficients and Chi-Squared (from solver members)
    iterations = solver.generations
    cost = solver.bestEnergy
    print("Generation %d has best Chi-Squared: %f" % (iterations, cost))
    print("Solved Coefficients:\n %s\n" % poly1d(solver.bestSolution))

    # compare solution with actual 8th-order Chebyshev coefficients
    print("Actual Coefficients:\n %s\n" % poly1d(chebyshev8coeffs))

    # plot solution versus exact coefficients
    plot_solution(solution)
    getch()

    # plot convergence of coefficients per iteration
    plot_frame('iterations')
    plot_params(stepmon)
    getch()

    # plot convergence of coefficients per function call
    plot_frame('function calls')
    plot_params(evalmon)
    getch()

# end of file



#!/usr/bin/env python
#
# Problem definition:
# Example in reference documentation for scipy.optimize
# http://docs.scipy.org/doc/scipy/reference/tutorial/optimize.html
# 
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

from optqp import _objective, bounds, xs, ys

from mystic.penalty import quadratic_equality
from mystic.constraints import with_penalty

@with_penalty(quadratic_equality, k=1e4)
def penalty(x): # == 0.0
    return x[0]**3 - x[1]


if __name__ == '__main__':

  from mystic.solvers import diffev2, fmin_powell
  from mystic.math import almostEqual

  result = diffev2(_objective, x0=bounds, bounds=bounds, penalty=penalty, npop=40, ftol=1e-8, gtol=100, disp=False, full_output=True)
  assert almostEqual(result[0], xs, rel=2e-2)
  assert almostEqual(result[1], ys, rel=2e-2)

  result = fmin_powell(_objective, x0=[-1.0,1.0], bounds=bounds, penalty=penalty, disp=False, full_output=True)
  assert almostEqual(result[0], xs, rel=2e-2)
  assert almostEqual(result[1], ys, rel=2e-2)


# EOD



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Author: Patrick Hung (patrickh @caltech)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
convert bound instances into functions
"""
# from dejong import sphere, rosen, step, quartic, shekel
def sphere(x):
    from mystic.models.dejong import sphere; return sphere(x)
from .dejong import sphere as model
sphere.__doc__ = model.__doc__

def rosen(x):
    from mystic.models.dejong import rosen; return rosen(x)
from .dejong import rosen as model
rosen.__doc__ = model.__doc__

def rosen0der(x):
    from mystic.models.dejong import rosen0der; return rosen0der(x)
from .dejong import rosen0der as model
rosen0der.__doc__ = model.__doc__

def rosen1der(x):
    from mystic.models.dejong import rosen1der; return rosen1der(x)
from .dejong import rosen1der as model
rosen1der.__doc__ = model.__doc__

def step(x):
    from mystic.models.dejong import step; return step(x)
from .dejong import step as model
step.__doc__ = model.__doc__

def quartic(x):
    from mystic.models.dejong import quartic; return quartic(x)
from .dejong import quartic as model
quartic.__doc__ = model.__doc__

def shekel(x):
    from mystic.models.dejong import shekel; return shekel(x)
from .dejong import shekel as model
shekel.__doc__ = model.__doc__

# from storn import corana, griewangk, zimmermann
def corana(x):
    from mystic.models.storn import corana; return corana(x)
from .storn import corana as model
corana.__doc__ = model.__doc__

def griewangk(x):
    from mystic.models.storn import griewangk; return griewangk(x)
from .storn import griewangk as model
griewangk.__doc__ = model.__doc__

def zimmermann(x):
    from mystic.models.storn import zimmermann; return zimmermann(x)
from .storn import zimmermann as model
zimmermann.__doc__ = model.__doc__

# from wolfram import fosc3d, nmin51
def fosc3d(x):
    from mystic.models.wolfram import fosc3d; return fosc3d(x)
from .wolfram import fosc3d as model
fosc3d.__doc__ = model.__doc__

def nmin51(x):
    from mystic.models.wolfram import nmin51; return nmin51(x)
from .wolfram import nmin51 as model
nmin51.__doc__ = model.__doc__

# from nag import peaks
def peaks(x):
    from mystic.models.nag import peaks; return peaks(x)
from .nag import peaks as model
peaks.__doc__ = model.__doc__

# from venkataraman import venkat91
def venkat91(x):
    from mystic.models.venkataraman import venkat91; return venkat91(x)
from .venkataraman import venkat91 as model
venkat91.__doc__ = model.__doc__

# from wavy import wavy1, wavy2
def wavy1(x):
    from mystic.models.wavy import wavy1; return wavy1(x)
from .wavy import wavy1 as model
wavy1.__doc__ = model.__doc__

def wavy2(x):
    from mystic.models.wavy import wavy2; return wavy2(x)
from .wavy import wavy2 as model
wavy2.__doc__ = model.__doc__

# from pohlheim import schwefel, ellipsoid, rastrigin, powers, ackley
# from pohlheim import michal, branins, easom, goldstein
def schwefel(x):
    from mystic.models.pohlheim import schwefel; return schwefel(x)
from .pohlheim import schwefel as model
schwefel.__doc__ = model.__doc__

def ellipsoid(x):
    from mystic.models.pohlheim import ellipsoid; return ellipsoid(x)
from .pohlheim import ellipsoid as model
ellipsoid.__doc__ = model.__doc__

def rastrigin(x):
    from mystic.models.pohlheim import rastrigin; return rastrigin(x)
from .pohlheim import rastrigin as model
rastrigin.__doc__ = model.__doc__

def powers(x):
    from mystic.models.pohlheim import powers; return powers(x)
from .pohlheim import powers as model
powers.__doc__ = model.__doc__

def ackley(x):
    from mystic.models.pohlheim import ackley; return ackley(x)
from .pohlheim import ackley as model
ackley.__doc__ = model.__doc__

def michal(x):
    from mystic.models.pohlheim import michal; return michal(x)
from .pohlheim import michal as model
michal.__doc__ = model.__doc__

def branins(x):
    from mystic.models.pohlheim import branins; return branins(x)
from .pohlheim import branins as model
branins.__doc__ = model.__doc__

def easom(x):
    from mystic.models.pohlheim import easom; return easom(x)
from .pohlheim import easom as model
easom.__doc__ = model.__doc__

def goldstein(x):
    from mystic.models.pohlheim import goldstein; return goldstein(x)
from .pohlheim import goldstein as model
goldstein.__doc__ = model.__doc__

# from schittkowski import paviani
def paviani(x):
    from mystic.models.schittkowski import paviani; return paviani(x)
from .schittkowski import paviani as model
paviani.__doc__ = model.__doc__

# clean up
del model


# EOF



#!/usr/bin/env python
#
# Problem definition:
# Example in reference documentation for cvxopt
# http://cvxopt.org/examples/tutorial/lp.html
# 
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
    Minimize: f = 2*x[0] + 1*x[1]

    Subject to:  -1*x[0] + 1*x[1] <= 1
                  1*x[0] + 1*x[1] >= 2
                           1*x[1] >= 0
                  1*x[0] - 2*x[1] <= 4

    where:  -inf <= x[0] <= inf
"""

def objective(x):
  x0,x1 = x
  return 2*x0 + x1

equations = """
-x0 + x1 - 1.0 <= 0.0
-x0 - x1 + 2.0 <= 0.0
x0 - 2*x1 - 4.0 <= 0.0
"""
bounds = [(None, None),(0.0, None)]

# with penalty='penalty' applied, solution is:
xs = [0.5, 1.5]
ys = 2.5

from mystic.symbolic import generate_conditions, generate_penalty
pf = generate_penalty(generate_conditions(equations), k=1e3)
from mystic.symbolic import generate_constraint, generate_solvers, simplify
cf = generate_constraint(generate_solvers(simplify(equations)))



if __name__ == '__main__':

  from mystic.solvers import diffev2, fmin_powell
  from mystic.math import almostEqual

  result = diffev2(objective, x0=bounds, bounds=bounds, penalty=pf, constraint=cf, npop=40, disp=False, full_output=True, ftol=1e-10, gtol=100)
  assert almostEqual(result[0], xs, rel=1e-2)
  assert almostEqual(result[1], ys, rel=1e-2)

  result = fmin_powell(objective, x0=[0.0,0.0], bounds=bounds, penalty=pf, constraint=cf, disp=False, full_output=True, gtol=3)
  assert almostEqual(result[0], xs, rel=1e-2)
  assert almostEqual(result[1], ys, rel=1e-2)


# EOF



#!/usr/bin/env python
#
# Problem definition and original response:
# https://stackoverflow.com/q/48088516/2379433
# https://stackoverflow.com/a/48494431/2379433
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2018-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Maximize:

  sum_{i=1 to max i} P_i O_i


Subject to:

  O_i <= C_i
  P_i <= 0


where the decision variables are:

  P_i: price allocated for night i


and these are the computed auxiliary variables:

  X_a,L: number of rooms allocated for stay of type (a,L)


defined as:

  X_a,L = d_a,L * ([sum_{i=a to a+L-i} P_i]/[L * P_nominal])^e


and:

  O_i: number of rooms reserved in a given night


defined as:

  O_i = sum_{a,L in N_i} X_a,L


The input parameters are:

  P_nominal: nominal price of the hotel (average historical price)

  e: elasticity between price and demand

  d_a,L and N_i: as defined in the classical model

  C_i: total number of rooms available on the hotel

"""
import math
n_days = 7
n_rooms = 50
P_nom = 85
P_bounds = 0,None
P_elastic = 2

class hotel(object):
    def __init__(self, rooms, price_ave, price_elastic):
        self.rooms = rooms
        self.price_ave = price_ave
        self.price_elastic = price_elastic

    def profit(self, P):
        # assert len(P) == len(self.rooms)
        return sum(j * self._reserved(P, i) for i,j in enumerate(P))

    def remaining(self, P): # >= 0
        C = self.rooms
        # assert len(P) == C
        return [C[i] - self._reserved(P, i) for i,j in enumerate(P)]

    def _reserved(self, P, day):
        max_days = len(self.rooms)
        As = range(0, day)
        return sum(self._allocated(P, a, L) for a in As
                   for L in range(day-a+1, max_days+1))
        
    def _allocated(self, P, a, L):
        P_nom = self.price_ave
        e = self.price_elastic
        return math.ceil(self._demand(a, L)*(sum(P[a:a+L])/(P_nom*L))**e)

    def _demand(self, a,L):
        return abs(1-a)/L + 2*(a**2)/L**2


h = hotel([n_rooms]*n_days, P_nom, P_elastic)

def objective(price, hotel):
    return -hotel.profit(price) 

def constraint(price, hotel): # <= 0
    return -min(hotel.remaining(price))

bounds = [P_bounds]*n_days
guess = [P_nom]*n_days

import mystic as my

@my.penalty.quadratic_inequality(constraint, kwds=dict(hotel=h))
def penalty(x):
    return 0.0

# using a local optimizer, starting from the nominal price
solver = my.solvers.fmin
mon = my.monitors.VerboseMonitor(50)

kwds = dict(disp=True, full_output=True, itermon=mon,
            args=(h,),  xtol=1e-8, ftol=1e-8, maxfun=10000, maxiter=2000)
result = solver(objective, guess, bounds=bounds, penalty=penalty, **kwds)

print([round(i,2) for i in result[0]])


# however, we can do better using a global optimizer
solver = my.solvers.diffev
mon = my.monitors.VerboseMonitor(50)

kwds = dict(disp=True, full_output=True, itermon=mon, npop=40,
            args=(h,),  gtol=250, ftol=1e-8, maxfun=30000, maxiter=2000)
result = solver(objective, bounds, bounds=bounds, penalty=penalty, **kwds)

print([round(i,2) for i in result[0]])



#!/usr/bin/env python
#
# Problem definition:
# A-R Hedar and M Fukushima, "Derivative-Free Filter Simulated Annealing
# Method for Constrained Continuous Global Optimization", Journal of
# Global Optimization, 35(4), 521-549 (2006).
# 
# Original Matlab code written by A. Hedar (Nov. 23, 2005)
# http://www-optima.amp.i.kyoto-u.ac.jp/member/student/hedar/Hedar_files/go.htm
# and ported to Python by Mike McKerns (December 2014)
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

def objective(x):
    x0,x1,x2,x3,x4,x5,x6 = x
    return (x0-10)**2 + 5*(x1-12)**2 + x2**4 + 3*(x3-11)**2 + \
           10*x4**6 + 7*x5**2 + x6**4 - 4*x5*x6 - 10*x5 - 8*x6

bounds = [(-10.,10.)]*7
# with penalty='penalty' applied, solution is:
xs = [2.330499, 1.951372, -0.4775414, 4.365726, -0.6244870, 1.038131, 1.594227]
ys = 680.6300573

from mystic.symbolic import generate_constraint, generate_solvers, simplify
from mystic.symbolic import generate_penalty, generate_conditions

equations = """
2.0*x0**2 + 3.0*x1**4 + x2 + 4.0*x3**2 + 5.0*x4 - 127.0 <= 0.0
7.0*x0 + 3.0*x1 + 10.0*x2**2 + x3 - x4 - 282.0 <= 0.0
23.0*x0 + x1**2 + 6.0*x5**2 - 8.0*x6 - 196.0 <= 0.0
4.0*x0**2 + x1**2 - 3.0*x0*x1 + 2.0*x2**2 + 5.0*x5 - 11.0*x6 <= 0.0
"""
cf = generate_constraint(generate_solvers(simplify(equations)))
pf = generate_penalty(generate_conditions(equations), k=1e12)



if __name__ == '__main__':

    from mystic.solvers import diffev2
    from mystic.math import almostEqual

    result = diffev2(objective, x0=bounds, bounds=bounds, constraints=cf, penalty=pf, npop=40, gtol=200, disp=False, full_output=True)

    assert almostEqual(result[0], xs, rel=1e-2)
    assert almostEqual(result[1], ys, rel=1e-2)



# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
#
# Adapted from seesaw2d.py in branches/UQ/math/examples2/ 
# For usage example, see seesaw2d_inf_example.py .
"""
TESTS for Dirac measure data objects.
Includes point_mass, measure, and product_measure classes.
"""

from mystic.math.discrete import point_mass as point
from mystic.math.discrete import measure as set
from mystic.math.discrete import product_measure as collection
from mystic.math.samples import random_samples
from mystic.math.grid import samplepts

from mystic.math import almostEqual
from mystic.math.measures import *
from mystic.math.measures import _pack, _unpack

disp = False

def test_calculate_methods(npts=2):
  upper_bounds = [1.0]
  lower_bounds = [0.0]

  # -------------------------------------
  # generate initial coordinates, weights 
  # -------------------------------------
  # get a random distribution of points
  if disp: print("generate random points and weights")
  coordinates = samplepts(lower_bounds, upper_bounds, npts)
  D0 = D = [i[0] for i in coordinates]
  if disp: print("positions: %s" % D)

  # calculate sample range
  R0 = R = spread(D)
  if disp: print("range: %s" % R)

  # select weights randomly in [0,1], then normalize so sum(weights) = 1
  wts = random_samples([0],[1], npts)[0]
  weights = normalize(wts, 0.0, zsum=True)
  if disp: print("weights (when normalized to 0.0): %s" % weights)
  assert almostEqual(sum(weights), 0.0, tol=1e-15)
  weights = normalize(wts, 1.0)
  assert almostEqual(sum(weights), 1.0, tol=1e-15)
  if disp: print("weights (when normalized to 1.0): %s" % weights)
  w = norm(weights)
  if disp: print("norm: %s" % w)
  assert almostEqual(w, sum(weights)/npts)

  # calculate sample mean
  m0 = m = mean(D,weights)
  if disp: print("mean: %s" % m)
  if disp: print("")

  # -------------------------------------
  # modify coordinates, maintaining mean & range 
  # -------------------------------------
  # get new random distribution
  if disp: print("modify positions, maintaining mean and range")
  coordinates = samplepts(lower_bounds, upper_bounds, npts)
  D = [i[0] for i in coordinates]

  # impose a range and mean on the points
  D = impose_spread(R, D, weights)
  D = impose_mean(m, D, weights)

  # print results
  if disp: print("positions: %s" % D)
  R = spread(D)
  if disp: print("range: %s" % R)
  assert almostEqual(R, R0)
  m = mean(D, weights)
  if disp: print("mean: %s" % m)
  assert almostEqual(m, m0)
  if disp: print("")

  # -------------------------------------
  # modify weights, maintaining mean & norm
  # -------------------------------------
  # select weights randomly in [0,1]
  if disp: print("modify weights, maintaining mean and range")
  wts = random_samples([0],[1], npts)[0]

  # print intermediate results
 #print("weights: %s" % wts)
 #sm = mean(D, wts)
 #print("tmp mean: %s" % sm)
 #print("")

  # impose mean and weight norm on the points
  D = impose_mean(m, D, wts)
  DD, weights = impose_weight_norm(D, wts)

  # print results
  if disp: print("weights: %s" % weights)
  w = norm(weights)
  if disp: print("norm: %s" % w)
  assert almostEqual(w, sum(weights)/npts)
  if disp: print("positions: %s" % DD)
  R = spread(DD)
  if disp: print("range: %s" % R)
  assert almostEqual(R, R0)
  sm = mean(DD, weights)
  if disp: print("mean: %s" % sm)
  assert almostEqual(sm, m0)
  sv = variance(DD, weights)
  if disp: print("var: %s" % sv)
  assert not almostEqual(sv, R)
  assert almostEqual(sv, 0.0, tol=.3)

  # -------------------------------------
  # modify variance, maintaining mean
  # -------------------------------------
  if disp: print("\nmodify variance, maintaining mean")
  DD = impose_variance(R, DD, weights)
  sm = mean(DD, weights)
  if disp: print("mean: %s" % sm)
  assert almostEqual(sm, m0)
  sv = variance(DD, weights)
  if disp: print("var: %s" % sv)
  assert almostEqual(sv, R)
  

def test_set_behavior():
  from mystic.math import almostEqual
  from numpy import inf

  # check basic behavior for set of two points
  s = set([point(1.0,1.0), point(3.0,2.0)]) #XXX: s + [pt, pt] => broken
  assert almostEqual(s.mean, 2.33333333)
  assert almostEqual(s.range, 2.0)
  assert almostEqual(s.mass, 3.0)

  # basic behavior for an admissable set
  s.normalize()
  s.mean = 1.0
  s.range = 1.0
  assert almostEqual(s.mean, 1.0)
  assert almostEqual(s.range, 1.0)
  assert almostEqual(s.mass, 1.0)

  # add and remove points
  # test special cases: SUM(weights)=0, RANGE(samples)=0, SUM(samples)=0
  s.append(point(1.0,-1.0))
  assert s.mean == -inf
  assert almostEqual(s.range, 1.0)
  assert almostEqual(s.mass, 0.0)

  '''
  _ave = s.mean
  s.mean = 1.0
  assert str(s.mean) == 'nan'
  assert str(s.range) == 'nan'
  assert almostEqual(s.mass, 0.0)
  s.normalize()
  assert str(s.mass) == 'nan'
  s.mean = _ave
  '''

  s.pop()
  s[0] = point(1.0,1.0)
  s[1] = point(-1.0,1.0)
  assert almostEqual(s.mean, 0.0)
  assert almostEqual(s.range, 2.0)
  assert almostEqual(s.mass, 2.0)
  s.normalize()
  s.mean = 1.0
  assert almostEqual(s.mean, 1.0)
  assert almostEqual(s.range, 2.0)
  assert almostEqual(s.mass, 1.0)

  s[0] = point(1.0,1.0)
  s[1] = point(1.0,1.0)
  assert almostEqual(s.mean, 1.0)
  assert almostEqual(s.range, 0.0)
  assert almostEqual(s.mass, 2.0)

  s.range = 1.0
  assert str(s.mean) == 'nan'
  assert str(s.range) == 'nan'
  assert almostEqual(s.mass, 2.0)
  return


def test_pack_unpack():
  x = [[1,2,3],[4,5],[6,7,8,9]]
  n = [len(i) for i in x]
  assert x == _unpack(_pack(x),n)
  return

def test_collection_behavior():
  from mystic.math import almostEqual
  from numpy import inf
  def f(x): return sum(x)  # a test function for expectation value

  # build three sets (x,y,z)
  sx = set([point(1.0,1.0), point(2.0,1.0), point(3.0,2.0)])
  sy = set([point(0.0,1.0), point(3.0,2.0)])
  sz = set([point(1.0,1.0), point(2.0,3.0)])
  #NOTE: for marc_surr(x,y,z), we must have x > 0.0
  sx.normalize()
  sy.normalize()
  sz.normalize()
  assert sx.mass == sy.mass == sz.mass == 1.0

  # build a collection
  c = collection([sx,sy,sz])
  xpos = c[0].positions
  ypos = c[1].positions
  zpos = c[2].positions
  xwts = c[0].weights
  ywts = c[1].weights
  zwts = c[2].weights
  if disp:
    print("x_positions: %s" % xpos)
    print("y_positions: %s" % ypos)
    print("z_positions: %s" % zpos)
    print("x_weights: %s" % xwts)
    print("y_weights: %s" % ywts)
    print("z_weights: %s" % zwts)
  assert xpos == sx.positions
  assert ypos == sy.positions
  assert zpos == sz.positions
  assert xwts == sx.weights
  assert ywts == sy.weights
  assert zwts == sz.weights
  tol = .2
  supp = c.support(tol)
  positions = c.positions
  weights = c.weights
  assert supp == [p for (p,w) in zip(positions,weights) if w > tol]
  if disp:
    print("support points:\n %s" % supp)
    print("npts: %s (i.e. %s)" % (c.npts, c.pts))
    print("weights: %s" % weights)
    print("positions: %s" % positions)
  assert c.npts == sx.npts * sy.npts * sz.npts
  assert len(weights) == len(positions) == c.npts
  assert sx.positions in c.pos
  assert sy.positions in c.pos
  assert sz.positions in c.pos
  assert sx.weights in c.wts
  assert sy.weights in c.wts
  assert sz.weights in c.wts

  c_exp = c.expect(f)
  if disp:
    print("mass: %s" % c.mass)
    print("expect: %s" % c_exp)
  assert c.mass == [sx.mass, sy.mass, sz.mass]
  assert c_exp == expectation(f, c.positions, c.weights)

 #print("center: %s" % c.center)
 #print("delta: %s" % c.delta)

  # change the positions in the collection
  points = [ list(i) for i in positions[::3] ]
  for i in range(len(points)):
    points[i][0] = 0.5
  positions[::3] = points
  c.positions = positions

  _xpos = c[0].positions
  _ypos = c[1].positions
  _zpos = c[2].positions
  _cexp = c.expect(f)
  if disp:
    print("x_positions: %s" % _xpos)
    print("y_positions: %s" % _ypos)
    print("z_positions: %s" % _zpos)
    print("expect: %s" % _cexp)
  assert _xpos == [0.5] + xpos[1:]
  assert _ypos == ypos
  assert _zpos == zpos
  assert _cexp < c_exp # due to _xpos[0] is 0.5 and less than 1.0

  _mean = 85.0
  _range = 0.25

  c.set_expect(_mean, f, npop=40, maxiter=200, tol=_range)
  _exp = c.expect(f)
  if disp:
    print("mean: %s" % _mean)
    print("range: %s" % _range)
    print("expect: %s" % _exp)
  assert almostEqual(_mean, _exp, tol=_mean*0.01)

  # a test function for probability of failure
  def g(x):
    if f(x) <= 0.0: return False
    return True
  pof = c.pof(g)
  spof = c.sampled_pof(g, npts=10000)
  if disp:
    print("pof: %s" % pof)
    print("sampled_pof: %s" % spof)
  assert almostEqual(pof, spof, tol=0.02)
  return

def test_flatten_unflatten():
  # build three sets (x,y,z)
  sx = set([point(1.0,1.0), point(2.0,1.0), point(3.0,2.0)])
  sy = set([point(0.0,1.0), point(3.0,2.0)])
  sz = set([point(1.0,1.0), point(2.0,3.0)])

  # build a collection
  c = collection([sx,sy,sz])

  # flatten and unflatten
  from mystic.math.discrete import flatten, unflatten
  d = unflatten(flatten(c), c.pts)

  # check if the same
  assert c.npts == d.npts
  assert c.weights == d.weights
  assert c.positions == d.positions

  # flatten() and load(...)
  e = collection().load(c.flatten(), c.pts)

  # check if the same
  assert c.npts == e.npts
  assert c.weights == e.weights
  assert c.positions == e.positions

  # decompose and compose
  from mystic.math.discrete import decompose, compose
  b = compose(*decompose(c))

  # check if the same
  assert c.npts == b.npts
  assert c.weights == b.weights
  assert c.positions == b.positions
  return


def test_min_max():
  # build a collection
  c = collection().load([.1, .4, 0, .2, .3, 1, 2, 6, 3, 5], (5,))

  # build a function
  f = lambda x: sum((i*i - 2*i) for i in x)

  # check min, max, etc
  assert c.minimum(f) == -1 == minimum(f, c.positions)
  assert c.maximum(f) == 24 == maximum(f, c.positions)
  assert c.ptp(f) == 25 == ptp(f, c.positions)
  assert c.ess_minimum(f) == -1 == ess_minimum(f, c.positions, c.weights)
  assert c.ess_maximum(f) == 15 == ess_maximum(f, c.positions, c.weights)
  assert c.ess_ptp(f) == 16 == ess_ptp(f, c.positions, c.weights)
  assert c.sampled_minimum(f, 100) == -1
  assert c.sampled_maximum(f, 100) == 15
  assert c.sampled_ptp(f, 100) == 16
  return


if __name__ == '__main__':
  test_calculate_methods(npts=2)
  test_set_behavior()
  test_pack_unpack()
  test_collection_behavior()
  test_flatten_unflatten()
  test_min_max()


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @uqfoundation)
# Copyright (c) 2020-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
functions to add noise to function inputs or outputs
"""

import numpy as np

def _noisy(x, dist=None, op=None):
  """make x noisy, by sampling from dist; returns op(x, dist.rvs(x.shape))

  x: parameters (list or array)
  dist: mystic.math.Distribution object
  op: numpy operator (e.g. np.add or np.prod)

  For example:
    >>> from numpy.random import normal
    >>> import mystic
    >>> x = [1,2,3]
    >>> rng = mystic.random_state('numpy.random', new=True, seed=123)
    >>> dist = mystic.math.Distribution(normal, 0, .1, rng=rng)
    >>> noisy(x, dist)
    [0.8914369396699439, 2.0997345446583586, 3.0282978498051993]
  """
  if op is None: op = np.add
  _type = type(x) if type(x) is not np.ndarray else None
  x = np.asarray(x)
  if dist is None:
    from mystic.math import Distribution
    dist = Distribution()
  dx = dist.rvs(x.shape)
  return op(x,dx) if _type is None else _type(op(x,dx))


def noisy(x, mu=0, sigma=1, seed='!'):
  """make x noisy, by adding noise from a normal distribution

  x: parameters (list or array)
  mu: distribution mean value
  sigma: distribution standard deviation
  seed: random seed [default: '!', do not reseed the RNG]
  """ # new: if True, generate a new RNG object [default: use the global RNG]
  from numpy.random import normal
  import mystic
  rng = mystic.random_state('numpy.random', new=True, seed=seed)
  dist = mystic.math.Distribution(normal, mu, sigma, rng=rng)
  return _noisy(x, dist=dist)




#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

import mystic.collapse as ct
import mystic.mask as ma
import mystic as my
m = my.monitors._load('_log.py')
# cleanup *pyc
import os
try: os.remove('_log.pyc')
except OSError: pass

import mystic.termination as mt
from mystic.solvers import DifferentialEvolutionSolver
solver = DifferentialEvolutionSolver(2*sum(m._npts))
solver.SetRandomInitialPoints()
solver.SetGenerationMonitor(m)
##############################################

# print('collapse_at')
ix = ct.collapse_at(m)
# (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 16, 17)
assert not ct.collapse_at(m, mask=ix)
ix = ct.collapse_at(m, target=0.0)
# (1, 7, 8, 9, 10, 11, 12, 14)
assert not ct.collapse_at(m, target=0.0, mask=ix)
ix = ct.collapse_at(m, target=m.x[-1]) 
assert not ct.collapse_at(m, target=m.x[-1], mask=ix)

# print('collapse_as')
ix = ct.collapse_as(m)
# set([(10, 11), (7, 12), (10, 12), (8, 9), (11, 14), (7, 11), (1, 11), (16, 17), (8, 14), (1, 14), (8, 10), (9, 11), (7, 10), (1, 10), (7, 14), (9, 14), (12, 14), (8, 11), (9, 10), (1, 9), (11, 12), (7, 9), (1, 12), (8, 12), (3, 4), (1, 8), (10, 14), (6, 13), (1, 7), (7, 8), (9, 12)])
assert not ct.collapse_as(m, mask=ix)

# print('collapse_weight')
ix = ct.collapse_weight(m)
# {0: {1}, 1: {1, 2}, 2: {0, 2}}
assert not ct.collapse_weight(m, mask=ix)

# print('collapse_position')
ix = ct.collapse_position(m)
# {0: {(0,1)}, 1:{(0,1),(0,2),(1,2)}, 2:{(1,2)}}
assert not ct.collapse_position(m, mask=ix)

##############################################
# print(mt.CollapseAt()(solver, True))
# print(mt.CollapseAt(target=0.0)(solver, True))
# print(mt.CollapseAs()(solver, True))
# print(mt.CollapseWeight()(solver, True))
# print(mt.CollapseWeight(mask=set())(solver, True))
# print(mt.CollapseWeight(mask=())(solver, True))
# print(mt.CollapsePosition()(solver, True))
# print(mt.CollapsePosition(mask=set())(solver, True))
# print(mt.CollapsePosition(mask=())(solver, True))

def test_cc(termination, seed=None):
    term = termination(mask=seed)
    collapse = ct.collapsed(term(solver, True))
    mask = list(collapse.values())[-1]
    # full results as mask
    _term = termination(mask=mask)
    assert not ct.collapsed(_term(solver, True))
    # update mask with full results
    _term = ma.update_mask(term, collapse)
    assert mask == ma.get_mask(_term)
    assert not ct.collapsed(_term(solver, True))

for term in (mt.CollapseAt, mt.CollapseAs):
    test_cc(term)

for term in (mt.CollapseWeight, mt.CollapsePosition):
    for seed in (None, {}, set(), ()):
        test_cc(term, seed)

##############################################
stop = mt.Or((mt.ChangeOverGeneration(), \
              mt.CollapseWeight(), \
              mt.CollapsePosition()))

message = stop(solver, True)

_stop = ma.update_mask(stop, ct.collapsed(message))

message = _stop(solver, True)
assert message
collapse = ct.collapsed(message)
assert not collapse
stop_ = ma.update_mask(_stop, collapse)
assert mt.state(stop_) == mt.state(_stop)



# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

from mystic.math.measures import *
from mystic.math import almostEqual

def test_impose_reweighted_mean():

  x0 = [1,2,3,4,5]
  w0 = [3,1,1,1,1]
  m = 3.5

  w = impose_reweighted_mean(m, x0, w0)
  assert almostEqual(mean(x0,w), m)


def test_impose_reweighted_variance():

  x0 = [1,2,3,4,5]
  w0 = [3,1,1,1,1]
  v = 1.0

  w = impose_reweighted_variance(v, x0, w0)
  assert almostEqual(variance(x0,w), v)
  assert almostEqual(mean(x0,w0), mean(x0,w))


if __name__ == '__main__':
  test_impose_reweighted_mean()
  test_impose_reweighted_variance()


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

from mystic.math.measures import *
from mystic.math.discrete import *
from mystic.math import almostEqual
def f(x): return sum(x)/len(x)

def test_setexpect(m, expect=5.0, tol=0.001):
    #e_lo = expect - expect
    #e_hi = expect + expect
    m.set_expect(expect, f, tol=tol)#, bounds=([e_lo]*m.npts, [e_hi]*m.npts)) 
    assert almostEqual(m.expect(f), expect, tol=tol)

def test_setexpectvar(m, expect=5.0, tol=0.001):
    m.set_expect_var(expect, f, tol=tol)
    assert almostEqual(m.expect_var(f), expect, tol=tol)

def test_setexpectmeanvar(m, expect=5.0, tol=0.001):
    m.set_expect_mean_and_var((expect,expect), f, tol=tol)
    assert almostEqual(m.expect(f), expect, tol=tol)
    assert almostEqual(m.expect_var(f), expect, tol=tol)

def test_expect(m, f):
    pos = m.positions
    if not (len(pos) and isinstance(pos[0], tuple)): # then m is a measure
        pos = [[i] for i in pos]
    expect = expectation(f, pos, m.weights)
    assert almostEqual(m.expect(f), expect)

def test_expectvar(m, f):
    pos = m.positions
    if not (len(pos) and isinstance(pos[0], tuple)): # then m is a measure
        pos = [[i] for i in pos]
    expect = expected_variance(f, pos, m.weights)
    assert almostEqual(m.expect_var(f), expect)

def fail(f, val):
    try:
        f(val)
        raise AssertionError('should throw ValueError')
    except ValueError:
        pass
    except:
        raise AssertionError('should throw ValueError')


if __name__ == '__main__':
    d2 = measure([point_mass(1.0, 1.0), point_mass(3.0, 2.0)])
    d1 = measure([point_mass(3.0, 2.0)])
    d2b = measure([point_mass(2.0, 4.0), point_mass(4.0, 2.0)])
    p1 = product_measure([d1])
    p1b = product_measure([d2])
    p2 = product_measure([d2,d2b])
    p21 = product_measure([d2,d1])

    # expect for product_measure and measure
    test_expect(d2, f)
    test_expect(d2b, f)
    test_expect(p2, f)
    test_expectvar(d2, f)
    test_expectvar(d2b, f)
    test_expectvar(p2, f)

    # set_expect for measure
    test_setexpect(d2)
    test_setexpect(d2b)
    test_setexpectvar(d2)
    test_setexpectvar(d2b)
    test_setexpectmeanvar(d2)
    test_setexpectmeanvar(d2b)

    # set_expect for product_measure
    test_setexpect(p2)
    test_setexpectvar(p2)
    test_setexpectmeanvar(p2)

    # again, but for single-point measures
    # expect for product_measure and measure
    test_expect(d1, f)
    test_expect(p1, f)
    test_expect(p1b, f)
    test_expectvar(d1, f)
    test_expectvar(p1, f)
    test_expectvar(p1b, f)

    # set_expect for measure
    test_setexpect(d1)
    test_setexpectvar(d1, expect=0.0) 
    test_setexpectmeanvar(d1, expect=0.0) 

    # set_expect for product_measure
    test_setexpect(p1)
    test_setexpect(p1b)
    test_setexpectvar(p1, expect=0.0)
    test_setexpectvar(p1b)
    test_setexpectmeanvar(p1, expect=0.0)
    test_setexpectmeanvar(p1b)

    # again, but for mixed measures
    test_setexpect(p21)
    test_setexpectvar(p21)
    test_setexpectmeanvar(p21)

    # failure cases (variance w/ one point_mass == 0.0)
    fail(test_setexpectvar, d1) 
    fail(test_setexpectmeanvar, d1)
    fail(test_setexpectvar, p1)
    fail(test_setexpectmeanvar, p1)


# EOF



#!/usr/bin/env python
#
# Problem definition:
# Example in google/or-tools
# https://github.com/google/or-tools/blob/master/examples/python/crypta.py
# with Copyright 2010 Hakan Kjellerstrand hakank@bonetmail.com
# and disclamer as stated at the above reference link.
# 
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
  Cryptarithmetic puzzle in Google CP Solver.

  Prolog benchmark problem
  '''
  Name           : crypta.pl
  Title          : crypt-arithmetic
  Original Source: P. Van Hentenryck's book
  Adapted by     : Daniel Diaz - INRIA France
  Date           : September 1992

  Solve the operation:

     B A I J J A J I I A H F C F E B B J E A
   + D H F G A B C D I D B I F F A G F E J E
   -----------------------------------------
   = G J E G A C D D H F A F J B F I H E E F
  '''
"""

def objective(x):
    return 0.0

bounds = [(0,9)]*10 + [(0,1)]*2
# with penalty='penalty' applied, solution is:
xs = [1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 1, 0]
ys = 0.0

# constraints
equations = """
A + 10*E + 100*J + 1000*B + 10000*B + 100000*E + 1000000*F + E + 10*J + 100*E + 1000*F + 10000*G + 100000*A + 1000000*F - F - 10*E - 100*E - 1000*H - 10000*I - 100000*F - 1000000*B - 10000000*Y == 0
C + 10*F + 100*H + 1000*A + 10000*I + 100000*I + 1000000*J + F + 10*I + 100*B + 1000*D + 10000*I + 100000*D + 1000000*C + Y - J - 10*F - 100*A - 1000*F - 10000*H - 100000*D - 1000000*D - 10000000*Z == 0
A + 10*J + 100*J + 1000*I + 10000*A + 100000*B + B + 10*A + 100*G + 1000*F + 10000*H + 100000*D + Z - C - 10*A - 100*G - 1000*E - 10000*J - 100000*G == 0
"""
var = list('ABCDEFGHIJYZ')
# correct bounds for:
# B - 1 >= 0
# D - 1 >= 0
# G - 1 >= 0
bounds[1] = bounds[3] = bounds[6] = (1,9)
#NOTE: FOR A MORE DIFFICULT PROBLEM, COMMENT OUT THE FOLLOWING 2 LINES
bounds[-1] = (0,0)
bounds[-2] = (1,1)

from mystic.constraints import unique, near_integers

from mystic.symbolic import generate_constraint, generate_solvers, solve
from mystic.symbolic import generate_penalty, generate_conditions
pf = generate_penalty(generate_conditions(equations,var),k=1e-6)
cf = generate_constraint(generate_solvers(solve(equations,var)))

from mystic.penalty import quadratic_inequality, quadratic_equality

#@quadratic_equality(near_integers)
def penalty(x):
    return pf(x)
   #return 0.0

#XXX: better way to constrain the last 2 (0,1) differently than rest (0,9)?
from numpy import round, hstack, clip
def constraint(x):
    # x[0:10] in range(0,10); x[10:] in range(0,2)
    x = round(x).astype(int) # force round and convert type to int
    x0,x1 = x[:-2],x[-2:]
    x0 = clip(x0, 0,9)       #XXX: hack to impose bounds
    x1 = clip(x1, 0,1)       #XXX: hack to impose bounds
    x0 = unique(x0, list(range(0,10)))
    return hstack([x0, x1])


if __name__ == '__main__':

    from mystic.solvers import diffev2, lattice
    from mystic.math import almostEqual
    from mystic.monitors import Monitor, VerboseMonitor
    mon = VerboseMonitor(10)#,10)

   #result = lattice(objective, 12, 960, bounds=bounds, penalty=penalty, constraints=constraint, ftol=1e-6, xtol=1e-6, disp=True, full_output=True, itermon=mon, maxiter=25)
   #result = diffev2(objective, x0=bounds, bounds=bounds, penalty=pf, constraints=constraint, npop=80, gtol=100, disp=True, full_output=True)
   #result = diffev2(objective, x0=bounds, bounds=bounds, penalty=penalty, constraints=constraint, npop=1000, ftol=1e-8, gtol=50, disp=True, full_output=True, cross=0.9, scale=0.9, itermon=mon)
    #FIXME: SOLVES at about 20%?... but w/ last 2 fixed 90%?
    result = diffev2(objective, x0=bounds, bounds=bounds, penalty=penalty, constraints=constraint, npop=360, ftol=1e-8, gtol=200, disp=True, full_output=True, cross=0.2, scale=0.9, itermon=mon)

    print(result[0])
    assert almostEqual(result[0], xs, tol=1e-8) #XXX: fails b/c rel & zero?
    assert almostEqual(result[1], ys, tol=1e-4)


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

import mystic.math.measures as mo
from numpy import nan, isnan
x = [1,2,3,4,5]
y = [1,2,3,4,5,6]
z = [1,5,5,5,5,5]
w = [1,2,3,4,10]

for i in (x,y,z,w):
    assert mo.moment(i, order=1) == 0.0
    assert mo.impose_moment(0.0, i, order=1) == i
    assert sum(isnan(mo.impose_moment(5.0, i, order=1))) == len(i)
    assert mo.impose_moment(0.0, i, order=2) == [mo.mean(i)]*len(i)
    assert sum(isnan(mo.impose_moment(5.0, i, order=2))) == 0
    assert mo.impose_moment(0.0, i, order=3) == [mo.mean(i)]*len(i)
    assert sum(isnan(mo.impose_moment(5.0, i, order=3))) == 0

assert sum(isnan(mo.impose_moment(5.0, x, order=3, skew=False))) == len(x)
assert sum(isnan(mo.impose_moment(5.0, y, order=3, skew=False))) == 0 #XXX?
assert sum(isnan(mo.impose_moment(5.0, z, order=3, skew=False))) == 0
assert sum(isnan(mo.impose_moment(5.0, w, order=3, skew=False))) == 0

for i in (x,y,z,w):
    tol = 1e-12
    _i = [max(i)+min(i)-j for j in i]
    assert mo.moment(i, order=3) + mo.moment(_i, order=3) <= tol
    assert mo.moment(i, order=5) + mo.moment(_i, order=5) <= tol

for i in (x,y,z,w):
    t = 5.0
    q = mo.impose_moment(t, i, order=3)
    s = mo.moment(q, order=3)
    assert round(t, 12) == round(s, 12)
    t = -t
    q = mo.impose_moment(t, i, order=3)
    s = mo.moment(q, order=3)
    assert round(t, 12) == round(s, 12)

for i in (x,y,z,w):
    t = 5.0
    q = mo.impose_moment(t, i, order=2)
    s = mo.moment(q, order=2)
    assert round(t, 12) == round(s, 12)
    t = -t
    q = mo.impose_moment(t, i, order=2)
    assert sum(isnan(q)) == len(i)




#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Sets up Corana's parabola. This is problem 6 of testbed 1 in [1].

Exact answer: Min = 0 @ abs(x_j) < 0.05 for all j.

Reference:

[1] Storn, R. and Price, K. Differential Evolution - A Simple and Efficient
Heuristic for Global Optimization over Continuous Spaces. Journal of Global
Optimization 11: 341-359, 1997.
"""

from mystic.solvers import DifferentialEvolutionSolver
from mystic.termination import ChangeOverGeneration, VTR
from mystic.strategy import Best1Exp, Rand1Exp

from mystic.tools import random_seed
random_seed(123)

from mystic.models import corana
from mystic.models.storn import Corana as Corana1
corana1 = Corana1(1)

ND = 4
NP = 10
MAX_GENERATIONS = 2500

def main():
    solver = DifferentialEvolutionSolver(ND, NP)

    solver.SetRandomInitialPoints(min = [-1000]*ND, max = [1000]*ND)
    solver.SetEvaluationLimits(generations=MAX_GENERATIONS)

    solver.Solve(corana, termination=VTR(0.00000001), strategy=Rand1Exp,\
                 CrossProbability=0.5, ScalingFactor=0.9)

    solution = solver.Solution()
  
    print(solution)



if __name__ == '__main__':
    from timeit import Timer
    t = Timer("main()", "from __main__ import main")
    timetaken =  t.timeit(number=1)
    print("CPU Time: %s" % timetaken)

    try:
        from mystic.solvers import fmin
       #from mystic._scipyoptimize import fmin
        import random
        print( "\nScipy: ")
        sol = fmin(corana, [random.random() for j in range(4)], full_output=0, retall=1)
        print("solution: %s" % sol[-1][0])
        print("\nCorana 1 with Scipy")
        sol = fmin(corana1, [random.random()], full_output=1, retall=1)
        print("solution: %s" % sol[-1][0])
    except:
        pass

# end of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Author: Patrick Hung (patrickh @caltech)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
models: sample models and functions prepared for use in mystic


Functions
=========

Mystic provides a set of standard fitting functions that derive from
the function API found in ``mystic.models.abstract_model``. These standard
functions are provided::

    sphere     -- De Jong's spherical function
    rosen      -- Sum of Rosenbrock's function
    step       -- De Jong's step function
    quartic    -- De Jong's quartic function
    shekel     -- Shekel's function
    corana     -- Corana's function
    fosc3d     -- Trott's fOsc3D function
    nmin51     -- Champion's NMinimize test function
    griewangk  -- Griewangk's function
    zimmermann -- Zimmermann's function
    peaks      -- NAG's peaks function
    venkat91   -- Venkataraman's sinc function
    schwefel   -- Schwefel's function
    ellipsoid  -- Pohlheim's rotated hyper-ellipsoid function
    rastrigin  -- Rastrigin's function
    powers     -- Pohlheim's sum of different powers function
    ackley     -- Ackley's path function
    michal     -- Michalewicz's function
    branins    -- Branins's rcos function
    easom      -- Easom's function
    goldstein  -- Goldstein-Price's function
    paviani    -- Paviani's function
    wavy1      -- a simple sine-based multi-minima function
    wavy2      -- another simple sine-based multi-minima function


Models
======

Mystic also provides a set of example models that derive from the model API
found in ``mystic.models.abstract_model``. These standard models are provided::

    poly       -- 1d model representation for polynomials
    circle     -- 2d array representation of a circle
    lorentzian -- Lorentzian peak model
    br8        -- Bevington & Robinson's model of dual exponential decay
    mogi       -- Mogi's model of surface displacements from a point spherical
                  source in an elastic half space

Additionally, ``circle`` has been extended to provide three additional models,
each with different packing densities::

    dense_circle, sparse_circle, and minimal_circle

Further, ``poly`` provides additional models for 2nd, 4th, 6th, 8th, and 16th
order Chebyshev polynomials::

    chebyshev2, chebyshev4, chebyshev6, chebyshev8, chebyshev16

Also, ``rosen`` has been modified to provide models for the 0th and 1st
derivative of the Rosenbrock function::

    rosen0der, and rosen1der
"""
# base classes
from .abstract_model import AbstractModel, AbstractFunction

# models
from .poly import poly, chebyshev2,chebyshev4,chebyshev6,chebyshev8,chebyshev16
from .mogi import mogi
from .br8 import decay
from .lorentzian import lorentzian
from .circle import circle, dense_circle, sparse_circle, minimal_circle

# functions
from .functions import *
# from dejong import sphere, rosen, step, quartic, shekel
# from storn import corana, griewangk, zimmermann
# from wolfram import fosc3d, nmin51
# from nag import peaks
# from venkataraman import venkat91
# from wavy import wavy1, wavy2
# from pohlheim import schwefel, ellipsoid, rastrigin, powers, ackley
# from pohlheim import michal, branins, easom, goldstein
# from schittkowski import paviani

#shortcuts
#from poly import chebyshev8cost, chebyshev16cost
#from br8 import data as br8data
#from br8 import cost as br8cost
#from lorentzian import gendata, histogram

# end of file
